import operator
from datetime import timedelta
from functools import reduce
import re
from itertools import chain
from typing import Union, List, Optional

import parsy
from parsy import string, Parser

from resotolib.parse_util import lexeme, float_p, integer_p


# See https://en.wikipedia.org/wiki/Unit_of_time for reference
# The order is relevant: from highest to lowest and longest to shortest
# | output | all names | number of seconds |
time_units = [
    ("yr", ["years", "year", "yr", "y"], 365 * 24 * 3600),
    ("mo", ["months", "month", "mo", "M"], 31 * 24 * 3600),
    ("d", ["days", "day", "d"], 24 * 3600),
    (None, ["weeks", "week", "w"], 7 * 24 * 3600),  # output is none, so it will not be used to print
    ("h", ["hours", "hour", "h"], 3600),
    ("min", ["minutes", "minute", "min", "m"], 60),
    ("s", ["seconds", "second", "s"], 1),
]

time_unit_combines = [",", "and"]

# Check if a string is a valid
DurationRe = re.compile(
    "^[+-]?([\\d.]+\\s*("
    + "|".join(chain.from_iterable(names for unit, names, _ in time_units))
    + ")\\s*("
    + "|".join(time_unit_combines)
    + ")?\\s*)+$"
)


def combine_durations(elems: List[Union[int, float]]) -> Union[int, float]:
    result = 0.0
    for d in elems:
        result += abs(d)
    return result if elems[0] >= 0 else -result


time_unit_parser = reduce(
    lambda x, y: x | y, [lexeme(string(name)).result(seconds) for _, names, seconds in time_units for name in names]
)

time_unit_combination: Parser = reduce(lambda x, y: x | y, [lexeme(string(a)) for a in [",", "and"]])
single_duration_parser = parsy.seq((float_p | integer_p), time_unit_parser).combine(operator.mul)
duration_parser = single_duration_parser.sep_by(time_unit_combination.optional(), min=1).map(combine_durations)


def parse_duration(ds: str) -> timedelta:
    return timedelta(seconds=duration_parser.parse(ds))


def parse_optional_duration(ds: str) -> Optional[timedelta]:
    try:
        return parse_duration(ds)
    except Exception:
        return None


def duration_str(duration: timedelta, precision: Optional[int] = 0, down_to_unit: Optional[str] = None) -> str:
    """
    Convert a timedelta to a string representing the duration human-readable short unit syntax.
    Examples: 3d2h, 2min42s
    :param duration: the duration to convert
    :param precision: the number of units to use to represent the duration.
    :param down_to_unit: use all units starting from the biggest one until the given unit.
    :return: string representing the duration
    """
    seconds = duration.total_seconds()
    found = False
    count = 0
    result = ""
    for unit, _, factor in time_units:
        if unit:
            if seconds >= factor:
                found = True
                num = int(seconds / factor)
                seconds = seconds - (num * factor)
                result += f"{num}{unit}"
            if found:
                count += 1
            if precision and count >= precision or unit == down_to_unit:
                break

    # in case the duration is less than one second
    return result if result else ("0s" if down_to_unit is None else f"0{down_to_unit}")



import pint
from pint.facets.plain import PlainUnit, PlainQuantity

reg = pint.UnitRegistry()

# Xi bytes are not known to pint
reg.define("Ei = 1 EiB")  # type: ignore
reg.define("Pi = 1 PiB")  # type: ignore
reg.define("Ti = 1 TiB")  # type: ignore
reg.define("Gi = 1 GiB")  # type: ignore
reg.define("Mi = 1 MiB")  # type: ignore
reg.define("Ki = 1 KiB")  # type: ignore
reg.define("KB = 1000 B")  # type: ignore

# globally define or register units

bytes_u: PlainUnit = reg.byte


def parse(s: str) -> PlainQuantity[PlainUnit]:
    return reg.parse_expression(s)



from datetime import datetime

import pytest

from resotolib.baseresources import InstanceStatus
from resotolib.json_bender import (
    StringToUnitNumber,
    bend,
    CPUCoresToNumber,
    MapValue,
    AsDate,
    Sort,
    S,
    MapEnum,
    StripNones,
    MapDict,
    F,
)


def test_map_value() -> None:
    assert bend(MapValue(dict(a=1, b=2)), "a") == 1
    assert bend(MapValue(dict(a=1, b=2)), "b") == 2
    assert bend(MapValue(dict(a=1, b=2)), "c") is None
    assert bend(MapValue(dict(a=1, b=2), default=3), "c") == 3


def test_map_enum() -> None:
    assert bend(MapEnum(dict(foo=InstanceStatus.BUSY)), "foo") == InstanceStatus.BUSY.value
    assert bend(MapEnum(dict(), InstanceStatus.BUSY), "foo") == InstanceStatus.BUSY.value
    with pytest.raises(AttributeError):
        bend(MapEnum(dict(foo="bla")), "foo")


def test_string_to_unit() -> None:
    assert bend(StringToUnitNumber("B", int), "4041564Ki") == 4138561536

    assert bend(StringToUnitNumber("GB", int), "4041564Ki") == 4
    assert bend(StringToUnitNumber("MB", int), "4041564Ki") == 4138
    assert bend(StringToUnitNumber("KB", int), "4041564Ki") == 4138561
    assert bend(StringToUnitNumber("GiB", int), "4041564Ki") == 3
    assert bend(StringToUnitNumber("MiB", int), "4041564Ki") == 3946
    assert bend(StringToUnitNumber("KiB", int), "4041564Ki") == 4041564

    assert bend(StringToUnitNumber("GiB", int), "2GiB") == 2
    assert bend(StringToUnitNumber("GiB", int), "2048MiB") == 2
    assert bend(StringToUnitNumber("GiB", int), "2097152KiB") == 2
    assert bend(StringToUnitNumber("GiB", int), "2Gi") == 2
    assert bend(StringToUnitNumber("GiB", int), "2048Mi") == 2
    assert bend(StringToUnitNumber("GiB", int), "2097152Ki") == 2


def test_cpu_cores_to_number() -> None:
    assert bend(CPUCoresToNumber(), "1") == 1
    assert bend(CPUCoresToNumber(), 1) == 1
    assert bend(CPUCoresToNumber(), "1000m") == 1
    assert bend(CPUCoresToNumber(), "3500m") == 3.5


def test_as_date() -> None:
    assert bend(AsDate(), "2022-06-07T14:43:49Z") == datetime(2022, 6, 7, 14, 43, 49)
    assert bend(AsDate(), None) is None


def test_sort() -> None:
    assert bend(Sort(S("i")), [{"i": 2}, {"i": 3}, {"i": 1}]) == [{"i": 1}, {"i": 2}, {"i": 3}]


def test_or_else() -> None:
    assert bend(S("a").or_else(S("b")), {"a": 1, "b": 2}) == 1
    assert bend(S("a").or_else(S("b")), {"b": 2}) == 2


def test_strip_nones() -> None:
    assert bend(StripNones(), [1, None, 2, None, None, 3]) == [1, 2, 3]


def test_map_dict() -> None:
    src = {"a": 1, "b": 2}
    assert bend(MapDict(value_bender=F(lambda x: x + 1)), src) == {"a": 2, "b": 3}
    assert bend(MapDict(key_bender=F(lambda x: x + "b")), src) == {"ab": 1, "bb": 2}
    assert bend(MapDict(key_bender=F(lambda x: x + "b"), value_bender=F(lambda x: x + 1)), src) == {"ab": 2, "bb": 3}



from attrs import define
from typing import Optional, ClassVar, Union

from resotolib.json import to_json, from_json, to_json_str


@define
class Foo:
    static: ClassVar[str] = "static"
    a: str
    b: int
    c: Optional[str] = None
    d: Union[str, int] = "test"
    inner: Optional["Foo"] = None
    _private: str = "private"


def test_roundtrip() -> None:
    foo = Foo("foo", 42, "bar")
    assert to_json(foo) == {"a": "foo", "b": 42, "c": "bar", "d": "test"}
    json = to_json(foo)
    assert json == {"a": "foo", "b": 42, "c": "bar", "d": "test"}
    again = from_json(json, Foo)
    assert isinstance(again, Foo)
    assert again == foo


def test_json_str() -> None:
    foo = Foo("foo", 42, "bar")
    assert to_json_str(foo) == '{"a": "foo", "b": 42, "c": "bar", "d": "test"}'
    assert to_json_str(foo, json_kwargs={"indent": 2}, strip_attr=("static", "a", "b", "c")) == '{\n  "d": "test"\n}'



from attrs import define
from enum import Enum

from resotolib.types import Json


class Severity(Enum):
    debug = "debug"
    info = "info"
    warn = "warn"
    error = "error"
    critical = "critical"


@define
class Event:
    # who created this message
    origin: str
    # seconds since epoch
    epoch: int
    # severity of this message
    severity: Severity
    # defines the type of the payload (e.g. log)
    kind: str
    # the payload of specific kind
    payload: Json



import os
import cherrypy
from prometheus_client.exposition import generate_latest, CONTENT_TYPE_LATEST


class WebApp:
    def __init__(self, mountpoint: str = "/") -> None:
        self.mountpoint = mountpoint
        local_path = os.path.abspath(os.path.dirname(__file__))
        config = {
            "tools.gzip.on": True,
            "tools.staticdir.index": "index.html",
            "tools.staticdir.on": True,
            "tools.staticdir.dir": f"{local_path}/static",
        }
        self.config = {"/": config}
        if self.mountpoint not in ("/", ""):
            self.config[self.mountpoint] = config

    @cherrypy.expose
    @cherrypy.tools.allow(methods=["GET"])
    def health(self) -> str:
        cherrypy.response.headers["Content-Type"] = "text/plain"
        return "ok\r\n"

    @cherrypy.expose
    @cherrypy.tools.allow(methods=["GET"])
    def metrics(self) -> bytes:
        cherrypy.response.headers["Content-Type"] = CONTENT_TYPE_LATEST
        return generate_latest()



from typing import ClassVar
from attrs import define
from resotolib.baseresources import BaseResource


@define(eq=False, slots=False)
class SomeTestResource(BaseResource):
    kind: ClassVar[str] = "some_test_resource"

    def delete(self, graph) -> bool:
        return False


def test_resource():
    r = SomeTestResource(id="foo", tags={"foo": "bar"})
    log_msg = "some log"
    r.log(log_msg)
    assert r.id == "foo"
    assert r.name == "foo"
    assert r.kind == "some_test_resource"
    assert r.tags["foo"] == "bar"
    assert r.event_log[0]["msg"] == log_msg



import re
from attrs import define
from typing import List, Dict, Tuple, Type, Optional, Any

from resotolib.baseresources import BaseResource
from resotolib.core.model_export import dataclasses_to_resotocore_model
from resotolib.json import from_json
from resotolib.types import Json


@define
class CheckProp:
    name: str
    kind: str


@define
class CheckClass:
    fqn: str
    aggregate_root: bool
    bases: Optional[List[str]]
    properties: List[CheckProp]

    def ignore(self) -> bool:
        # Those types were created during development of 2.4 and renamed. They were never available in a final release.
        # In case somebody operated on edge, we want to ignore them.
        return self.fqn.startswith("aws_auto_scaling") or self.fqn.startswith("aws_quota")


def check_overlap_for(models: List[Json]) -> None:
    # make sure that all model names are unique
    all_fqns = set()
    for model in models:
        if model["fqn"] in all_fqns:
            raise Exception(f"Model {model['fqn']} is defined multiple times")
        all_fqns.add(model["fqn"])
    # convert json representation to intermediate python structure
    classes = {model["fqn"]: from_json(model, CheckClass) for model in models if "properties" in model}
    # this variable holds all possible property paths
    all_paths: Dict[str, Tuple[CheckClass, str]] = {}

    # checks if 2 kinds are compatible
    def is_compatible(left: str, right: str) -> bool:
        return left == "any" or right == "any" or left == right

    def add_path(path: List[str], kinds: List[CheckClass], model: CheckClass) -> None:
        # This check is required to prevent endless loops: consider class Foo with property inner of type Foo.
        # We would walk this chain infinitely, that's why we return as early as possible
        for c in kinds:
            if c == model:
                return

        # Walk all properties of the model and add them to the all_paths dict.
        for prop in model.properties:
            # add the current kind to the list: this must be a new list which is unique for the property path
            pkinds = kinds + [model]
            kind = prop.kind
            prop_path = path + [prop.name]
            if "[]" in prop.kind:
                kind = prop.kind.replace("[]", "")
                prop_path += ["[0]"]  # use always the first element for simplicity
            elif "dictionary[" in prop.kind:
                kind = re.sub("dictionary\\[[^,]+,\\s*(\\S*)\\s*]", r"\1", prop.kind)
                prop_path += ["foo"]  # always use foo as lookup key

            # Create a string representation of the path. E.g. user.address.city.zip
            str_path = ".".join(prop_path)

            # Check if the path is already in the list of all paths and has a compatible kind.
            if existing := all_paths.get(str_path):
                existing_class, existing_kind = existing
                if not is_compatible(existing_kind, prop.kind):
                    raise AttributeError(
                        f"{str_path} is defined in {existing_class.fqn} as {existing_kind} and in {model.fqn} as {kind}"
                    )

            # update the dict of all paths, ignoring any existing value
            all_paths[str_path] = (model, prop.kind)

            # if this property  kind is complex too: walk it.
            if check_kind := classes.get(kind):
                add_path(prop_path, pkinds, check_kind)

    # Walk all models and add all properties to the all_paths dict.
    for _, clazz in classes.items():
        if clazz.aggregate_root and not clazz.ignore():
            add_path([], [], clazz)

    # Check that all successor kinds exist
    for model in models:
        if succ_dict := model.get("successor_kinds"):
            for kinds in succ_dict.values():
                for kind in kinds:
                    if kind not in all_fqns:
                        raise AttributeError(f"Successor kind {kind} does not exist")


def check_overlap(*base: Type[BaseResource]) -> None:
    """
    Call this method from your collector plugin to check for overlapping properties.
    This will try to load all models from all known plugins.
    The call will fail if the imports are not working - make sure the calling side has all those plugins installed.

    @param base: additional base classes to check for overlapping properties. All existing known plugins are added.
    :raise Exception: if there is an overlap
    """

    def dynamic_import(name: str) -> List[Type[Any]]:
        components = name.split(".")
        mod = __import__(components[0])
        for comp in components[1:]:
            mod = getattr(mod, comp)
        if isinstance(mod, type):
            return [mod]
        elif isinstance(mod, list):
            return mod
        else:
            raise AttributeError(f"Import {name}: expected type or list of types, got {type(mod)}")

    # List of all plugin classes that need to be imported.
    model_classes = {
        *dynamic_import("resoto_plugin_aws.collector.all_resources"),
        *dynamic_import("resoto_plugin_digitalocean.resources.DigitalOceanResource"),
        *dynamic_import("resoto_plugin_dockerhub.resources.DockerHubResource"),
        *dynamic_import("resoto_plugin_example_collector.ExampleResource"),
        *dynamic_import("resoto_plugin_gcp.resources.base.GcpResource"),
        *dynamic_import("resoto_plugin_github.resources.GithubResource"),
        *dynamic_import("resoto_plugin_k8s.resources.KubernetesResource"),
        *dynamic_import("resoto_plugin_onelogin.OneLoginResource"),
        *dynamic_import("resoto_plugin_onprem.resources.OnpremResource"),
        *dynamic_import("resoto_plugin_posthog.resources.PosthogResource"),
        *dynamic_import("resoto_plugin_random.resources.RandomResource"),
        *dynamic_import("resoto_plugin_scarf.resources.ScarfResource"),
        *dynamic_import("resoto_plugin_slack.resources.SlackResource"),
        *dynamic_import("resoto_plugin_vsphere.resources.VSphereResource"),
        *base,
    }
    # check overlap for all plugin classes
    check_overlap_for(dataclasses_to_resotocore_model(model_classes, aggregate_root=BaseResource))



from typing import List, Dict

import pytest

from resotolib.config import Config, current_config
from resotolib.core.custom_command import (
    execute_command,
    execute_command_on_resource,
    command_definitions,
    CommandDefinition,
)
from resotolib.types import Json, JsonElement


class SomeExamplePlugin:
    @execute_command(
        name="some_name",
        info="some info",
        description="some description",
        args_description={"arg1": "some description", "arg2": "some description"},
        expect_node_result=True,
        filter={"cloud": ["aws", "azure"], "region": ["eu-west-1", "eu-west-2"]},
    )
    def some_name(self, config: Config, js: Json, args: List[str]) -> JsonElement:
        return {"js": js, "args": args}

    @execute_command_on_resource(
        name="some_other",
        info="some info",
        description="some description",
        args_description={"foo": "some description", "bar": "some description"},
    )
    def some_other(self, config: Config, js: Json, args: List[str]) -> JsonElement:
        raise AttributeError("some error")


def test_command_definitions() -> None:
    # The str class does not define any custom commands
    no_definitions = command_definitions(str)
    assert len(no_definitions) == 0

    # The SomeExamplePlugin class defines two custom commands
    definitions = command_definitions(SomeExamplePlugin)
    assert len(definitions) == 2
    def_by_name: Dict[str, CommandDefinition] = {d.name: d for d in definitions}
    some_name = def_by_name["some_name"]
    assert some_name.info == "some info"
    assert some_name.description == "some description"
    assert some_name.args_description == {"arg1": "some description", "arg2": "some description"}
    assert some_name.filter == {"cloud": ["aws", "azure"], "region": ["eu-west-1", "eu-west-2"]}
    assert some_name.expect_node_result is True
    assert some_name.expect_resource is False
    assert some_name.allowed_on_kind is None
    result = some_name.fn(SomeExamplePlugin(), current_config(), {"foo": "bar"}, ["arg1", "arg2"])
    assert result == {"js": {"foo": "bar"}, "args": ["arg1", "arg2"]}

    some_other = def_by_name["some_other"]
    with pytest.raises(AttributeError) as ex:
        some_other.fn(SomeExamplePlugin(), current_config(), {"foo": "bar"}, ["arg1", "arg2"])
    assert str(ex.value) == "some error"



from typing import Dict, Any, Union, Mapping, Sequence, TypeVar, Callable, Any

# mypy does not support recursive type definitions
# See discussion here: https://github.com/python/typing/issues/182
Json = Dict[str, Any]
JsonElement = Union[str, int, float, bool, None, Mapping[str, Any], Sequence[Any]]
DecoratedFn = TypeVar("DecoratedFn", bound=Callable[..., Any])



import sys
import select
import asyncio
from typing import AsyncIterator


async def async_lines(async_iterator: AsyncIterator[str]) -> AsyncIterator[str]:
    buffer = ""
    async for chunk in async_iterator:
        buffer += chunk
        lines = buffer.split("\n")
        for line in lines[:-1]:
            yield line.rstrip("\r\n")
        buffer = lines[-1]
    if buffer:
        yield buffer


async def stdin_generator() -> AsyncIterator[str]:
    has_data = await asyncio.to_thread(lambda: select.select([sys.stdin], [], [], 0.0)[0])

    if has_data:
        for line in iter(sys.stdin.readline, ""):
            yield line.rstrip("\r\n")



import json
import logging
import re
import time
from contextvars import ContextVar
from datetime import datetime, timedelta
from re import RegexFlag
from typing import Any, Dict, Optional, Set, Callable, Awaitable, Tuple
from urllib.parse import urlparse

from aiohttp import web
from aiohttp.web import Request, StreamResponse
from aiohttp.web import middleware
from attr import define
from jwt import PyJWTError

from resotolib import jwt as ck_jwt
from resotolib.asynchronous.web import RequestHandler, Middleware
from resotolib.jwt import encode_jwt
from resotolib.types import Json
from resotolib.utils import utc

log = logging.getLogger(__name__)
JWT = Dict[str, Any]
__JWT_Context: ContextVar[JWT] = ContextVar("JWT", default={})
CodeLifeTime = timedelta(minutes=5)


@define
class AuthorizedUser:
    email: str
    roles: Set[str]
    authorized_at: datetime

    def is_valid(self) -> bool:
        return utc() - self.authorized_at < CodeLifeTime


__authorization_codes: Dict[str, AuthorizedUser] = {}


async def authorized_user(code: str) -> Optional[AuthorizedUser]:
    for invalid_code in [k for k, v in __authorization_codes.items() if not v.is_valid()]:
        __authorization_codes.pop(invalid_code, None)
    return __authorization_codes.get(code)


def add_authorized_user(code: str, user: AuthorizedUser) -> None:
    __authorization_codes[code] = user


async def jwt_from_context() -> JWT:
    """
    Inside a request handler, this value retrieves the current jwt.
    """
    return __JWT_Context.get()


def raw_jwt_from_auth_message(msg: str) -> Optional[str]:
    """
    Expected message: json object with type kind="authorization" and a jwt field
    { "kind": "authorization", "jwt": "Bearer <jwt>" }
    """
    try:
        js = json.loads(msg)
        assert js.get("kind") == "authorization"
        return js.get("jwt")  # type: ignore
    except Exception:
        return None


@middleware
async def no_check(request: Request, handler: RequestHandler) -> StreamResponse:
    # all requests are authorized automatically
    request["authorized"] = True
    return await handler(request)


def set_valid_jwt(request: Request, jwt_raw: str, psk: str) -> Optional[JWT]:
    try:
        # note: the expiration is already checked by this function
        jwt = ck_jwt.decode_jwt_from_header_value(jwt_raw, psk)
    except PyJWTError:
        return None
    if jwt:
        request["authorized"] = True  # deferred check in websocket handler
        request["jwt"] = jwt
        __JWT_Context.set(jwt)
    return jwt


def renew_user_jwt(psk: str, jwt_lifetime: timedelta, user: AuthorizedUser) -> Tuple[str, Json]:
    exp = int(time.time() + jwt_lifetime.total_seconds())
    data = {"email": user.email, "roles": ",".join(user.roles), "exp": exp}
    return encode_jwt(data, psk, expire_in=int(jwt_lifetime.total_seconds())), data


def check_auth(
    psk: str,
    jwt_lifetime: timedelta,
    always_allowed_paths: Set[str],
    not_allowed: Optional[Callable[[Request], Awaitable[StreamResponse]]] = None,
) -> Middleware:
    def always_allowed(request: Request) -> bool:
        for path in always_allowed_paths:
            if re.fullmatch(path, request.path, RegexFlag.IGNORECASE):
                return True
        return False

    async def valid_jwt(request: Request) -> bool:
        auth_header = request.headers.get("Authorization") or request.cookies.get("resoto_authorization")
        if auth_header:
            # make sure origin and host match, so the request is valid
            origin: Optional[str] = urlparse(request.headers.get("Origin")).hostname  # type: ignore
            host: Optional[str] = request.headers.get("Host")
            if host is not None and origin is not None:
                if ":" in host:
                    host = host.split(":")[0]
                if origin.lower() != host.lower():
                    log.warning(f"Origin {origin} is not allowed in request from {request.remote} to {request.path}")
                    raise web.HTTPForbidden()

            # try to authorize the request, even if it is one of the always allowed paths
            authorized = set_valid_jwt(request, auth_header, psk) is not None
            return authorized
        return False

    async def valid_code(request: Request) -> bool:
        code = request.query.get("code")
        if code:
            if (user := await authorized_user(code)) and user.is_valid():
                data = {"email": user.email, "roles": ",".join(user.roles)}
                jwt = encode_jwt(data, psk, expire_in=int(jwt_lifetime.total_seconds()))
                request["send_auth_response_header"] = f"Bearer {jwt}"
                request["jwt"] = data
                return True
        return False

    @middleware
    async def valid_auth_handler(request: Request, handler: RequestHandler) -> StreamResponse:
        allowed = False
        if always_allowed(request):
            allowed = True
        elif request.headers.get("Authorization"):
            allowed = await valid_jwt(request)
        elif request.query.get("code"):
            allowed = await valid_code(request)
        if allowed:
            request["authorized"] = True
            return await handler(request)
        else:
            if not_allowed:
                return await not_allowed(request)
            else:
                raise web.HTTPUnauthorized()

    return valid_auth_handler


def auth_handler(
    psk: Optional[str],
    jwt_lifetime: timedelta,
    always_allowed_paths: Set[str],
    not_allowed: Optional[Callable[[Request], Awaitable[StreamResponse]]] = None,
) -> Middleware:
    if psk:
        log.info("Use JWT authentication with a pre shared key")
        return check_auth(psk, jwt_lifetime, always_allowed_paths, not_allowed)
    else:
        log.info("No authentication requested.")
        return no_check



import os
import time
import warnings
import requests
from ssl import create_default_context, SSLContext
from typing import Tuple, Optional, List, Dict, Union, Any
from resotolib.args import ArgumentParser
from resotolib.core import resotocore
from resotolib.x509 import (
    csr_to_bytes,
    load_cert_from_bytes,
    cert_fingerprint,
    gen_rsa_key,
    gen_csr,
    cert_to_bytes,
    load_cert_from_file,
    write_cert_to_file,
    write_key_to_file,
    key_to_bytes,
    load_key_from_bytes,
    load_key_from_file,
    write_ca_bundle,
)
from resotolib.jwt import decode_jwt_from_headers, encode_jwt_to_headers
from cryptography.x509.base import Certificate
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from tempfile import TemporaryDirectory
from threading import Lock, Event, Thread, Condition
from resotolib.logger import log
from resotolib.event import add_event_listener, Event as ResotoEvent, EventType
from datetime import datetime, timedelta
from jwt.exceptions import InvalidSignatureError


class FingerprintError(Exception):
    pass


class NoJWTError(Exception):
    pass


def get_ca_cert(resotocore_uri: Optional[str] = None, psk: Optional[str] = None) -> Certificate:
    if resotocore_uri is None:
        resotocore_uri = resotocore.http_uri
    if psk is None:
        psk = getattr(ArgumentParser.args, "psk", None)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        r = requests.get(f"{resotocore_uri}/ca/cert", verify=False)
        ca_cert = load_cert_from_bytes(r.content)
        if psk:
            # noinspection PyTypeChecker
            jwt = decode_jwt_from_headers(r.headers, psk)
            if jwt is None:
                raise NoJWTError("Failed to decode JWT")
            if jwt["sha256_fingerprint"] != cert_fingerprint(ca_cert):
                raise FingerprintError("Invalid Root CA certificate fingerprint")
        return ca_cert


def get_signed_cert(
    common_name: str,
    san_dns_names: Optional[List[str]] = None,
    san_ip_addresses: Optional[List[str]] = None,
    resotocore_uri: Optional[str] = None,
    psk: Optional[str] = None,
    ca_cert_path: Optional[str] = None,
    connect_to_ips: Optional[List[str]] = None,
) -> Tuple[RSAPrivateKey, Certificate]:
    if resotocore_uri is None:
        resotocore_uri = getattr(ArgumentParser.args, "resotocore_uri", None)
    if psk is None:
        psk = getattr(ArgumentParser.args, "psk", None)

    cert_key = gen_rsa_key()
    cert_csr = gen_csr(
        cert_key,
        common_name=common_name,
        connect_to_ips=connect_to_ips,
        san_dns_names=san_dns_names,
        san_ip_addresses=san_ip_addresses,
    )
    cert_csr_bytes = csr_to_bytes(cert_csr)
    headers: Dict[str, str] = {}
    if psk is not None:
        encode_jwt_to_headers(headers, {}, psk)
    request_kwargs: Dict[str, str] = {}
    if ca_cert_path is not None:
        request_kwargs["verify"] = ca_cert_path
    r = requests.post(f"{resotocore_uri}/ca/sign", cert_csr_bytes, headers=headers, **request_kwargs)  # type: ignore
    if r.status_code != 200:
        raise ValueError(f"Failed to get signed certificate: {r.text}")
    cert_bytes = r.content
    cert_crt = load_cert_from_bytes(cert_bytes)
    return cert_key, cert_crt


class TLSData:
    def __init__(
        self,
        common_name: str,
        san_dns_names: Optional[List[str]] = None,
        san_ip_addresses: Optional[List[str]] = None,
        tempdir: Optional[str] = None,
        resotocore_uri: Optional[str] = None,
        psk: Optional[str] = None,
        ca_only: bool = False,
        renew_before: timedelta = timedelta(days=1),
    ) -> None:
        self.common_name = common_name
        self.san_dns_names = san_dns_names
        self.san_ip_addresses = san_ip_addresses
        self.__ca_only = ca_only
        self.renew_before = renew_before
        self.__tempdir = TemporaryDirectory(prefix="resoto-cert-", dir=tempdir)
        if resotocore_uri is None:
            resotocore_uri = resotocore.http_uri
        self.__resotocore_uri = resotocore_uri
        self.__psk = psk
        self.__ca_cert: Optional[Certificate] = None
        self.__cert: Optional[Certificate] = None
        self.__key: Optional[RSAPrivateKey] = None
        self.__ca_cert_path = f"{self.__tempdir.name}/ca.crt"
        self.__cert_path = f"{self.__tempdir.name}/cert.crt"
        self.__key_path = f"{self.__tempdir.name}/cert.key"
        self.__load_lock = Lock()
        self.__loaded = Event()
        self.__exit = Condition()
        self.__watcher: Optional[Thread] = None

    def __enter__(self) -> "TLSData":
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.shutdown()

    def start(self) -> None:
        if self.__watcher is None:
            self.load()
            self.__watcher = Thread(target=self.__certificates_watcher, name="certificates_watcher")
            self.__watcher.start()
            add_event_listener(EventType.SHUTDOWN, self.shutdown, blocking=False)

    def shutdown(self, event: Optional[ResotoEvent] = None) -> None:
        if self.__watcher is not None:
            with self.__exit:
                self.__exit.notify()
            self.__watcher.join()
            self.__watcher = None

    def __getstate__(self) -> Dict[str, Any]:
        d = self.__dict__.copy()
        del d["_TLSData__load_lock"]
        del d["_TLSData__loaded"]
        del d["_TLSData__exit"]
        del d["_TLSData__ca_cert"]
        del d["_TLSData__cert"]
        del d["_TLSData__key"]
        del d["_TLSData__watcher"]
        d["__is_loaded"] = self.__loaded.is_set()
        if self.__loaded.is_set():
            d["__ca_cert_bytes"] = cert_to_bytes(self.__ca_cert)  # type: ignore
            d["__cert_bytes"] = cert_to_bytes(self.__cert)  # type: ignore
            d["__key_bytes"] = key_to_bytes(self.__key)  # type: ignore
        return d

    def __setstate__(self, d: Dict[str, Any]) -> None:
        d["_TLSData__load_lock"] = Lock()
        d["_TLSData__loaded"] = Event()
        d["_TLSData__exit"] = Condition()
        d["_TLSData__ca_cert"] = None
        d["_TLSData__cert"] = None
        d["_TLSData__key"] = None
        if d["__is_loaded"]:
            d["_TLSData__loaded"].set()
            d["_TLSData__ca_cert"] = load_cert_from_bytes(d["__ca_cert_bytes"])
            d["_TLSData__cert"] = load_cert_from_bytes(d["__cert_bytes"])
            d["_TLSData__key"] = load_key_from_bytes(d["__key_bytes"])
            del d["__ca_cert_bytes"]
            del d["__cert_bytes"]
            del d["__key_bytes"]
        del d["__is_loaded"]
        d["_TLSData__watcher"] = Thread(target=self.__certificates_watcher, name="certificates_watcher")
        self.__dict__.update(d)

    def reload(self) -> None:
        self.__loaded.clear()
        self.load()

    def __certificates_watcher(self) -> None:
        while True:
            with self.__exit:
                if self.__loaded.is_set():
                    for cert in (self.__ca_cert, self.__cert):
                        if (
                            isinstance(cert, Certificate)
                            and cert.not_valid_after < datetime.utcnow() - self.renew_before
                        ):
                            self.reload()
                            break
                    self.__refresh_files_on_disk()
                if self.__exit.wait(60):
                    break

    def __refresh_files_on_disk(self, refresh_every_sec: int = 10800) -> None:
        if not self.__loaded.is_set():
            return
        try:
            last_ca_cert_update = time.time() - os.path.getmtime(self.__ca_cert_path)
            last_cert_update = time.time() - os.path.getmtime(self.__cert_path)
            if last_ca_cert_update > refresh_every_sec or last_cert_update > refresh_every_sec:
                log.debug("Refreshing cert/key files on disk")
                write_ca_bundle(self.__ca_cert, self.__ca_cert_path, include_certifi=True)  # type: ignore
                write_cert_to_file(self.__cert, self.__cert_path)  # type: ignore
                write_key_to_file(self.__key, self.__key_path)  # type: ignore
        except FileNotFoundError:
            pass

    def load(self) -> None:
        with self.__load_lock:
            if getattr(ArgumentParser.args, "ca_cert", None) is not None:
                log.debug(f"Loading CA certificate from {ArgumentParser.args.ca_cert}")
                self.__ca_cert = load_cert_from_file(ArgumentParser.args.ca_cert)
            else:
                log.debug("Loading CA cert from core")
                try:
                    self.__ca_cert = get_ca_cert(resotocore_uri=self.__resotocore_uri, psk=self.__psk)
                except FingerprintError as e:
                    log.fatal(f"{e}, MITM attack?")
                    raise
                except InvalidSignatureError as e:
                    log.fatal(f"{e}, wrong PSK?")
                    raise
                except NoJWTError as e:
                    log.fatal(f"{e}, resotocore started without PSK?")
                    raise
                except Exception as e:
                    log.fatal(f"{e}")
                    raise
            log.debug(f"Writing CA cert {self.__ca_cert_path}")
            write_ca_bundle(self.__ca_cert, self.__ca_cert_path, include_certifi=True)
            if not self.__ca_only:
                if (
                    getattr(ArgumentParser.args, "cert", None) is not None
                    and getattr(ArgumentParser.args, "cert_key", None) is not None
                ):
                    log.debug(f"Loading certificate from {ArgumentParser.args.cert}")
                    self.__cert = load_cert_from_file(ArgumentParser.args.cert)
                    cert_key_pass = None
                    if getattr(ArgumentParser.args, "cert_key_pass", None) is not None:
                        cert_key_pass = ArgumentParser.args.cert_key_pass
                    log.debug(f"Loading key from {ArgumentParser.args.cert_key}")
                    self.__key = load_key_from_file(ArgumentParser.args.cert_key, passphrase=cert_key_pass)
                else:
                    log.debug("Requesting signed cert from core")
                    self.__key, self.__cert = get_signed_cert(
                        common_name=self.common_name,
                        san_dns_names=self.san_dns_names,
                        san_ip_addresses=self.san_ip_addresses,
                        resotocore_uri=self.__resotocore_uri,
                        psk=self.__psk,
                        ca_cert_path=self.ca_cert_path,
                    )
                log.debug(f"Writing signed cert/key {self.__cert_path}")
                write_cert_to_file(self.__cert, self.__cert_path)
                write_key_to_file(self.__key, self.__key_path)
            self.__loaded.set()

    @property
    def ca_cert(self) -> Certificate:
        if not os.path.isfile(self.__ca_cert_path):
            self.load()
        assert self.__ca_cert is not None, "CA cert is None"
        return self.__ca_cert

    @property
    def cert(self) -> Certificate:
        if not self.__loaded.is_set():
            self.load()
        assert self.__cert is not None, "Cert is None"
        return self.__cert

    @property
    def key(self) -> RSAPrivateKey:
        if not self.__loaded.is_set():
            self.load()
        assert self.__key is not None, "Key is None"
        return self.__key

    @property
    def ca_cert_path(self) -> str:
        if not os.path.isfile(self.__ca_cert_path):
            self.load()
        return self.__ca_cert_path

    @property
    def cert_path(self) -> str:
        if not self.__loaded.is_set():
            self.load()
        return self.__cert_path

    @property
    def key_path(self) -> str:
        if not self.__loaded.is_set():
            self.load()
        return self.__key_path

    @property
    def sslopt(self) -> Dict[str, str]:
        return {"ca_certs": self.ca_cert_path}

    @property
    def verify(self) -> Union[str, bool]:
        if getattr(ArgumentParser.args, "verify_certs", None) is not None:
            if ArgumentParser.args.verify_certs is False:
                return False
        return self.ca_cert_path

    @property
    def ssl_context(self) -> SSLContext:
        context = create_default_context()
        context.load_verify_locations(cafile=self.ca_cert_path)
        return context

    @staticmethod
    def add_args(arg_parser: ArgumentParser, ca_only: bool = False) -> None:
        arg_parser.add_argument(
            "--ca-cert",
            help="Path to custom CA certificate file",
            default=None,
            type=str,
            dest="ca_cert",
        )
        if not ca_only:
            arg_parser.add_argument(
                "--cert",
                help="Path to custom certificate file",
                default=None,
                type=str,
                dest="cert",
            )
            arg_parser.add_argument(
                "--cert-key",
                help="Path to custom certificate key file",
                default=None,
                type=str,
                dest="cert_key",
            )
            arg_parser.add_argument(
                "--cert-key-pass",
                help="Passphrase for certificate key file",
                default=None,
                type=str,
                dest="cert_key_pass",
            )
        arg_parser.add_argument(
            "--no-verify-certs",
            help="Turn off certificate verification",
            default=True,
            dest="verify_certs",
            action="store_false",
        )



from resotolib.core.tasks import CoreTaskHandler, CoreTaskResult
from .custom_command_test import SomeExamplePlugin
from resotolib.core.custom_command import command_definitions, CommandDefinition


def test_conversion() -> None:
    definitions = {a.name: a for a in command_definitions(SomeExamplePlugin)}
    some_name: CommandDefinition = definitions["some_name"]
    handler = CoreTaskHandler.from_definition(SomeExamplePlugin(), some_name)
    assert handler.core_json() == {
        "name": some_name.name,
        "info": some_name.info,
        "args_description": some_name.args_description,
        "description": some_name.description,
        "expect_node_result": some_name.expect_node_result,
        "allowed_on_kind": some_name.allowed_on_kind,
        "filter": some_name.filter,
    }


def test_matches() -> None:
    handler = CoreTaskHandler(
        name="test", info="", description="", handler=lambda x: None, filter={"a": ["b"], "b": ["c"]}
    )
    assert handler.matches({"task_name": "test", "attrs": {"a": "b", "b": "c", "some": "other"}})
    assert not handler.matches({"task_name": "test", "attrs": {"a": "b"}})
    assert not handler.matches({"task_name": "test", "attrs": {}})


def test_call() -> None:
    definitions = {a.name: a for a in command_definitions(SomeExamplePlugin)}
    handler = CoreTaskHandler.from_definition(SomeExamplePlugin, definitions["some_name"])
    res = handler.execute({"task_id": "test", "data": {"args": ["a", "b"], "node": {"a": "b"}}})
    assert res == CoreTaskResult(task_id="test", data={"js": {"a": "b"}, "args": ["a", "b"]})


def test_exception() -> None:
    definitions = {a.name: a for a in command_definitions(SomeExamplePlugin)}
    handler = CoreTaskHandler.from_definition(SomeExamplePlugin, definitions["some_other"])
    res = handler.execute({"task_id": "test", "data": {"args": ["a", "b"], "node": None}})
    assert res == CoreTaskResult(task_id="test", error="some error")



import time
from abc import ABC, abstractmethod
from enum import Enum, auto
from threading import Thread, current_thread
from typing import Dict, Optional, Set, Any

from prometheus_client import Counter

import resotolib.config
import resotolib.proc
from resotolib.args import ArgumentParser
from resotolib.baseresources import BaseResource, Cloud
from resotolib.config import Config
from resotolib.core import resotocore
from resotolib.core.actions import CoreActions
from resotolib.core.ca import TLSData
from resotolib.graph import Graph
from resotolib.logger import log
from resotolib.types import Json

# from multiprocessing import Process

metrics_unhandled_plugin_exceptions = Counter(
    "resoto_unhandled_plugin_exceptions_total",
    "Unhandled plugin exceptions",
    ["plugin"],
)


class PluginType(Enum):
    """Defines Plugin Type

    COLLECTOR is a cloud resource collector plugin that gets instantiated
    on each collect() run
    PERSISTENT is a persistent plugin that gets instantiated once upon startup
    """

    COLLECTOR = auto()
    ACTION = auto()
    PERSISTENT = auto()
    CLI = auto()
    POST_COLLECT = auto()


class BasePlugin(ABC, Thread):
    """A resoto Plugin is a thread that does some work.

    If the plugin_type is PluginType.COLLECTOR the Plugin gets instantiated each
    collect run.
    If the plugin_type is PluginType.PERSISTENT the Plugin gets instantiated upon
    startup and is expected to run forever. It may register to any events it's
    interested in and act upon them.

    Upon start the go() method is called. For COLLECTOR Plugins collect() is called.
    """

    plugin_type = PluginType.PERSISTENT

    def __init__(self) -> None:
        super().__init__()
        self.name = self.__class__.__name__
        self.finished = False

    def run(self) -> None:
        try:
            self.go()
        except Exception:
            metrics_unhandled_plugin_exceptions.labels(plugin=self.name).inc()
            log.exception(f"Caught unhandled plugin exception in {self.name}")
        else:
            self.finished = True

    @abstractmethod
    def go(self) -> None:
        """Do the Plugin work"""
        pass

    @staticmethod
    def add_args(arg_parser: ArgumentParser) -> None:
        """Adds Plugin specific arguments to the global arg parser"""
        pass

    @staticmethod
    def add_config(config: Config) -> None:
        """Adds Plugin specific config options"""
        pass


class BaseActionPlugin(ABC, Thread):
    plugin_type = PluginType.ACTION
    action: str = NotImplemented  # Name of the action this plugin implements

    def __init__(self, tls_data: Optional[TLSData] = None) -> None:
        super().__init__()
        self._args = ArgumentParser.args
        self._config = resotolib.config._config
        self.name = self.__class__.__name__
        self.finished = False
        self.timeout = Config.resotoworker.timeout
        self.wait_for_completion = True
        self.tls_data = tls_data

    @abstractmethod
    def do_action(self, data: Dict[str, Any]) -> None:
        """Perform an action"""
        pass

    def action_processor(self, message: Json) -> Optional[Json]:
        """Process incoming action messages"""
        if not isinstance(message, dict):
            log.error(f"Invalid message: {message}")
            return None
        kind = message.get("kind")
        message_type = message.get("message_type")
        data: Json = message.get("data", {})
        log.debug(f"Received message of kind {kind}, type {message_type}, data: {data}")
        if kind == "action":
            try:
                if message_type == self.action:
                    start_time = time.time()
                    self.do_action(data)
                    run_time = int(time.time() - start_time)
                    log.debug(f"{self.action} ran for {run_time} seconds")
                else:
                    raise ValueError(f"Unknown message type {message_type}")
            except Exception as e:
                log.exception(f"Failed to {message_type}: {e}")
                reply_kind = "action_error"
            else:
                reply_kind = "action_done"

            return {
                "kind": reply_kind,
                "message_type": message_type,
                "data": data,
            }
        else:
            return None

    def run(self) -> None:
        try:
            # ArgumentParser.args = self._args
            # resotolib.config._config = self._config
            # setup_logger("resotoworker")
            # resotolib.proc.initializer()
            current_thread().name = self.name
            if self.bootstrap():
                self.go()
        except Exception:
            metrics_unhandled_plugin_exceptions.labels(plugin=self.name).inc()
            log.exception(f"Caught unhandled plugin exception in {self.name}")
        else:
            self.finished = True

    @abstractmethod
    def bootstrap(self) -> bool:
        """Bootstrap the plugin.

        If bootstrapping is successful the plugin is ready to run.
        """
        pass

    def go(self) -> None:
        core_actions = CoreActions(
            identifier=f"{ArgumentParser.args.subscriber_id}-actions-{self.action}-{self.name}",
            resotocore_uri=resotocore.http_uri,
            resotocore_ws_uri=resotocore.ws_uri,
            actions={
                self.action: {
                    "timeout": self.timeout,
                    "wait_for_completion": self.wait_for_completion,
                },
            },
            message_processor=self.action_processor,
            tls_data=self.tls_data,
        )
        core_actions.start()
        core_actions.join()

    @staticmethod
    def add_args(arg_parser: ArgumentParser) -> None:
        """Adds Plugin specific arguments to the global arg parser"""
        pass

    @staticmethod
    def add_config(config: Config) -> None:
        """Adds Plugin specific config options"""
        pass


class BaseCollectorPlugin(BasePlugin):
    """A resoto Collector plugin is a thread that collects cloud resources.

    Whenever the thread is started the collect() method is run. The collect() method
    is expected to add cloud resources to self.graph. Cloud resources must inherit
    the BaseResource or one of the more specific resource types like BaseAccount,
    BaseInstance, BaseNetwork, BaseLoadBalancer, etc.

    When the collect() method finishes, the Collector will retrieve the
    Plugins Graph and append it to the global Graph.
    """

    plugin_type = PluginType.COLLECTOR  # Type of the Plugin
    cloud: str = NotImplemented  # Name of the cloud this plugin implements

    def __init__(self) -> None:
        super().__init__()
        self.name = str(self.cloud)
        cloud = Cloud(id=self.cloud)
        self.root = cloud
        self.graph = Graph(root=self.root)

    @abstractmethod
    def collect(self) -> None:
        """Collects all the Cloud Resources"""
        pass

    @staticmethod
    def auto_enableable() -> bool:
        """Should this collector be enabled by default?"""
        return False

    @staticmethod
    def update_tag(config: Config, resource: BaseResource, key: str, value: str) -> bool:
        """Update the tag of a resource"""
        return resource.update_tag(key, value)

    @staticmethod
    def delete_tag(config: Config, resource: BaseResource, key: str) -> bool:
        """Delete the tag of a resource"""
        return resource.delete_tag(key)

    @staticmethod
    def pre_cleanup(config: Config, resource: BaseResource, graph: Graph) -> bool:
        return resource.pre_cleanup(graph)

    @staticmethod
    def cleanup(config: Config, resource: BaseResource, graph: Graph) -> bool:
        return resource.cleanup(graph)  # type: ignore

    def go(self) -> None:
        self.collect()


class BasePostCollectPlugin(ABC):
    """A resoto Post Collect plugin is a thread that runs after collection is done.

    Whenever the thread is started the post_collect() method is run. The post_collect() method
    is expected take the graph and perform operations on it, e.g. add the external edges or
    enritch the graph in some other way.
    """

    plugin_type = PluginType.POST_COLLECT  # Type of the Plugin
    name: str = NotImplemented  # Name of the cloud this plugin implements
    activate_with: Set[str]  # List of clouds this plugin should be activated on

    def __init__(self) -> None:
        super().__init__()
        self.name = self.name

    @abstractmethod
    def post_collect(self, graph: Graph) -> None:
        """Process the collected graph"""
        pass

    @staticmethod
    def add_args(arg_parser: ArgumentParser) -> None:
        """Adds Plugin specific arguments to the global arg parser"""
        pass

    @staticmethod
    def add_config(config: Config) -> None:
        """Adds Plugin specific config options"""
        pass



import asyncio
import logging
import socket
from ssl import SSLContext
from typing import Optional, Union, Awaitable, Callable, Type

from aiohttp.abc import AbstractAccessLogger

# noinspection PyProtectedMember
from aiohttp.helpers import all_tasks
from aiohttp.log import access_logger

# noinspection PyProtectedMember
from aiohttp.web import HostSequence, _run_app, _cancel_tasks
from aiohttp.web_app import Application
from aiohttp.web_log import AccessLogger
from aiohttp.web_runner import GracefulExit


# This method is copied from aiohttp.web.run_app with one 2 additional steps:
# - it allows a callable that is executed on shutdown.
# - it does not swallow terminal exceptions
def run_app(
    app: Union[Application, Awaitable[Application]],
    on_shutdown: Callable[[], Awaitable[None]],
    *,
    host: Optional[Union[str, HostSequence]] = None,
    port: Optional[int] = None,
    path: Optional[str] = None,
    sock: Optional[socket.socket] = None,
    shutdown_timeout: float = 60.0,
    keepalive_timeout: float = 75.0,
    ssl_context: Optional[SSLContext] = None,
    print_cmd: Callable[..., None] = print,
    backlog: int = 128,
    access_log_class: Type[AbstractAccessLogger] = AccessLogger,
    access_log_format: str = AccessLogger.LOG_FORMAT,
    access_log: Optional[logging.Logger] = access_logger,
    handle_signals: bool = True,
    reuse_address: Optional[bool] = None,
    reuse_port: Optional[bool] = None,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> None:
    """Run an app locally"""
    if loop is None:
        loop = asyncio.new_event_loop()

    # Configure if and only if in debugging mode and using the default logger
    if loop.get_debug() and access_log and access_log.name == "aiohttp.access":
        if access_log.level == logging.NOTSET:
            access_log.setLevel(logging.DEBUG)
        if not access_log.hasHandlers():
            access_log.addHandler(logging.StreamHandler())

    main_task = loop.create_task(
        _run_app(
            app,
            host=host,
            port=port,
            path=path,
            sock=sock,
            shutdown_timeout=shutdown_timeout,
            keepalive_timeout=keepalive_timeout,
            ssl_context=ssl_context,
            print=print_cmd,
            backlog=backlog,
            access_log_class=access_log_class,
            access_log_format=access_log_format,
            access_log=access_log,
            handle_signals=handle_signals,
            reuse_address=reuse_address,
            reuse_port=reuse_port,
        )
    )

    try:
        asyncio.set_event_loop(loop)
        loop.run_until_complete(main_task)
    except (GracefulExit, KeyboardInterrupt) as ex:  # pragma: no cover
        raise ex
    finally:
        loop.run_until_complete(on_shutdown())
        _cancel_tasks({main_task}, loop)
        _cancel_tasks(all_tasks(loop), loop)
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()



import json
from attrs import define, field
from typing import ClassVar, Optional, Dict, Mapping, Any
import sys
import os
import logging
from logging import basicConfig, getLogger, DEBUG, INFO, WARNING, ERROR, CRITICAL, StreamHandler, Formatter, LogRecord


from resotolib.args import ArgumentParser
from resotolib.types import Json

DEBUG2 = DEBUG - 1
DEBUG3 = DEBUG - 2
DEBUG4 = DEBUG - 3
DEBUG5 = DEBUG - 4
TRACE = DEBUG - 5

getLogger().setLevel(ERROR)
getLogger("resoto").setLevel(INFO)


def add_args(arg_parser: ArgumentParser) -> None:
    group = arg_parser.add_mutually_exclusive_group()
    group.add_argument(
        "--verbose",
        "-v",
        help="Verbose logging",
        dest="verbose",
        action="store_true",
        default=False,
    )
    group.add_argument(
        "--trace",
        help="Trage logging",
        dest="trace",
        action="store_true",
        default=False,
    )
    group.add_argument(
        "--quiet",
        help="Only log errors",
        dest="quiet",
        action="store_true",
        default=False,
    )


@define
class LoggingConfig:
    kind: ClassVar[str] = "logging"
    verbose: Optional[bool] = field(default=False, metadata={"description": "Verbose logging"})
    quiet: Optional[bool] = field(default=False, metadata={"description": "Only log errors"})


class JsonFormatter(Formatter):
    """
    Simple json log formatter.
    Inspired by: https://stackoverflow.com/questions/50144628/python-logging-into-file-as-a-dictionary-or-json
    """

    def __init__(
        self,
        fmt_dict: Mapping[str, str],
        time_format: str = "%Y-%m-%dT%H:%M:%S",
        static_values: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__()
        self.fmt_dict = fmt_dict
        self.time_format = time_format
        self.static_values = static_values or {}
        self.__use_time = "asctime" in self.fmt_dict.values()

    def usesTime(self) -> bool:  # noqa: N802
        return self.__use_time

    def formatMessage(self, record: LogRecord) -> dict:  # type: ignore # noqa: N802
        return {fmt_key: record.__dict__[fmt_val] for fmt_key, fmt_val in self.fmt_dict.items()}

    def formatJsonMessage(self, record) -> Json:  # type: ignore # noqa: N802
        record.message = record.getMessage()

        if self.__use_time:
            record.asctime = self.formatTime(record, self.time_format)

        message_dict = self.formatMessage(record)
        message_dict.update(self.static_values)

        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)

        if record.exc_text:
            message_dict["exception"] = record.exc_text

        if record.stack_info:
            message_dict["stack_info"] = self.formatStack(record.stack_info)
        return message_dict

    def format(self, record: LogRecord) -> str:
        message_dict = self.formatJsonMessage(record)
        return json.dumps(message_dict, default=str)


def setup_logger(
    proc: str,
    *,
    force: bool = True,
    verbose: bool = False,
    quiet: bool = False,
    level: Optional[str] = None,
    json_format: bool = True,
) -> None:
    # override log output via env var
    plain_text = os.environ.get("RESOTO_LOG_TEXT", "false").lower() == "true"
    if json_format and not plain_text:
        handler = StreamHandler()
        formatter = JsonFormatter(
            {
                "timestamp": "asctime",
                "level": "levelname",
                "message": "message",
                "pid": "process",
                "thread": "threadName",
            },
            static_values={"process": proc},
        )
        handler.setFormatter(formatter)
        basicConfig(handlers=[handler], force=force, level=level)
    else:
        log_format = f"%(asctime)s|{proc}|%(levelname)5s|%(process)d|%(threadName)10s  %(message)s"
        # allow to define the log format via env var
        log_format = os.environ.get("RESOTO_LOG_FORMAT", log_format)
        basicConfig(format=log_format, datefmt="%y-%m-%d %H:%M:%S", force=force)
    argv = sys.argv[1:]
    if level:
        getLogger("resoto").setLevel(level)
    elif "--trace" in argv or os.environ.get("RESOTO_TRACE", "false").lower() == "true":
        getLogger("resoto").setLevel(TRACE)
    elif verbose or "-v" in argv or "--verbose" in argv or os.environ.get("RESOTO_VERBOSE", "false").lower() == "true":
        getLogger("resoto").setLevel(DEBUG)
    elif quiet or "--quiet" in argv or os.environ.get("RESOTO_QUIET", "false").lower() == "true":
        getLogger().setLevel(WARNING)
        getLogger("resoto").setLevel(CRITICAL)


# via https://stackoverflow.com/a/35804945/92184
def add_logging_level(level_name: str, level_num: int, method_name: Optional[str] = None) -> None:
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> add_logging_level("TRACE", logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace("that worked")
    >>> logging.trace("so did this")
    >>> logging.TRACE
    5

    """
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        raise AttributeError("{} already defined in logging module".format(level_name))
    if hasattr(logging, method_name):
        raise AttributeError("{} already defined in logging module".format(method_name))
    if hasattr(logging.getLoggerClass(), method_name):
        raise AttributeError("{} already defined in logger class".format(method_name))

    def log_for_level(self: logging.Logger, message: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message: str, *args: Any, **kwargs: Any) -> None:
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


add_logging_level("DEBUG2", DEBUG2)
add_logging_level("DEBUG3", DEBUG3)
add_logging_level("DEBUG4", DEBUG4)
add_logging_level("DEBUG5", DEBUG5)
add_logging_level("TRACE", TRACE)

setup_logger("resoto", force=False)
log = getLogger("resoto")



import threading
import cherrypy
from resotolib.logger import log
from typing import Optional, Any


class WebServer(threading.Thread):
    def __init__(
        self,
        web_app: Any,
        web_host: str = "::",
        web_port: int = 9955,
        ssl_cert: Optional[str] = None,
        ssl_key: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.name = "webserver"
        self.web_app = web_app
        self.web_host = web_host
        self.web_port = web_port
        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key

    @property
    def serving(self) -> bool:
        return bool(cherrypy.engine.state == cherrypy.engine.states.STARTED)

    def run(self) -> None:
        # CherryPy always prefixes its log messages with a timestamp.
        # The next line monkey patches that time method to return a
        # fixed string. So instead of having duplicate timestamps in
        # each web server related log message they are now prefixed
        # with the string 'CherryPy'.
        cherrypy.config.reset()
        cherrypy._cplogging.LogManager.time = lambda self: "CherryPy"
        cherrypy.engine.unsubscribe("graceful", cherrypy.log.reopen_files)

        # We always mount at / as well as any user configured --web-path
        cherrypy.tree.mount(
            self.web_app,
            "",
            self.web_app.config,
        )
        if self.web_app.mountpoint not in ("/", ""):
            cherrypy.tree.mount(
                self.web_app,
                self.web_app.mountpoint,
                self.web_app.config,
            )
        ssl_args = {}
        if self.ssl_cert and self.ssl_key:
            ssl_args = {
                "server.ssl_module": "builtin",
                "server.ssl_certificate": self.ssl_cert,
                "server.ssl_private_key": self.ssl_key,
            }
        cherrypy.config.update(
            {
                "global": {
                    "engine.autoreload.on": False,
                    "server.socket_host": self.web_host,
                    "server.socket_port": self.web_port,
                    "log.screen": False,
                    "log.access_file": "",
                    "log.error_file": "",
                    "tools.log_headers.on": False,
                    "tools.encode.on": True,
                    "tools.encode.encoding": "utf-8",
                    "request.show_tracebacks": False,
                    "request.show_mismatched_params": False,
                    **ssl_args,
                }
            }
        )
        cherrypy.engine.start()
        cherrypy.engine.block()

    def shutdown(self) -> None:
        log.debug("Received request to shutdown http server threads")
        cherrypy.engine.exit()



"""
Resoto common library
~~~~~~~~~~~~~~~~~~~~~~~~~~
Common Resoto functions.
:copyright: © 2023 Some Engineering Inc.
:license: Apache 2.0, see LICENSE for more details.
"""

__title__ = "resotolib"
__description__ = "Resoto common library."
__author__ = "Some Engineering Inc."
__license__ = "Apache 2.0"
__copyright__ = "Copyright © 2023 Some Engineering Inc."
__version__ = "3.5.0"



import re
from typing import Callable

import parsy
from parsy import Parser, generate, regex, string


def make_direct_parser(fn: Callable[[str, int], parsy.Result]) -> Parser:
    """
    Make typed parser (required for mypy).
    """
    return Parser(fn)


def make_parser(fn: Callable[[], Parser]) -> Parser:
    """
    Make typed parser (required for mypy).
    """
    return generate(fn)


whitespace: Parser = regex(r"\s*")


def lexeme(p: Parser) -> Parser:
    """
    Create a whitespace tolerant parser from a given parser.
    """
    return whitespace >> p << whitespace


# region direct parser (not including whitespace)

space_dp = regex(r"\s+")
lparen_dp = string("(")
rparen_dp = string(")")
l_bracket_dp = string("[")
r_bracket_dp = string("]")
l_curly_dp = string("{")
r_curly_dp = string("}")
gt_dp = string(">")
lt_dp = string("<")
slash = string("/")
colon_dp = string(":")
semicolon_dp = string(";")
comma_dp = string(",")
dot_dp = string(".")
dot_dot_dp = string("..")
equals_dp = string("=")
backtick_dp = string("`")
backslash_dp = string("\\")
pipe_dp = string("|")
dash_dp = string("-")
integer_dp = regex(r"[+-]?[0-9]+").map(int)
float_dp = regex(r"[+-]?[0-9]+\.[0-9]+").map(float)
literal_dp = regex("[A-Za-z0-9][A-Za-z0-9_\\-]*")

# variables:
# - simple props: foo, bla, bar
# - nested props: foo.bla.bar
# - array props: foo[*].bla[0].bar[1]
# - props with non json conform keys: foo.bla.`:-)`.bar
variable_dp_backtick_allowed = regex(r"[^`]+")
variable_dp_backtick = backtick_dp + variable_dp_backtick_allowed + backtick_dp
variable_dp_part_string = regex("[A-Za-z_][A-Za-z0-9_\\-]*")
variable_dp_part_plain = regex("[.]*")
variable_dp_part = variable_dp_part_string | variable_dp_backtick
variable_dp_array_part = variable_dp_part + regex("(\\[[0-9\\*]*\\])*")
optional_slash = slash.at_most(1).concat()
variable_dp = optional_slash + variable_dp_array_part + (dot_dp + variable_dp_array_part).many().map("".join)
variable_no_array_dp = optional_slash + (variable_dp_part + variable_dp_part_plain).at_least(1).map("".join)


unquoted_allowed_characters = re.compile("[A-Za-z0-9_\\-:/.]")
unquoted_end_of_unquoted_str = re.compile("[,\\[\\])(}{\\s]")


def unquoted_string_parser(*stop_words: str) -> Parser:
    @make_direct_parser
    def unquoted_string_direct_parser(stream: str, index: int) -> parsy.Result:
        # read from index until -> end_of_unquoted_str
        # all characters in between have to be allowed characters
        # there has to be at least one non-digit character (to not read numbers as strings)
        start = index
        found_no_number = False
        valid_string_chars = True
        number_found_dot = False

        # valid numbers are: digits or a minus at the start: 123, -321, -123.321
        def is_no_number() -> bool:
            nonlocal number_found_dot
            char = stream[index]
            is_dot = char == "."
            is_num = str.isdigit(char) or (start == index and char == "-") or (is_dot and not number_found_dot)
            if not number_found_dot and is_dot:
                number_found_dot = True
            return not is_num

        # look ahead to next delimiter
        while index < len(stream) and valid_string_chars and not unquoted_end_of_unquoted_str.match(stream[index]):
            found_no_number = is_no_number() if not found_no_number else found_no_number
            valid_string_chars = unquoted_allowed_characters.match(stream[index]) is not None
            index += 1

        result = stream[start:index]
        for stop in stop_words:
            stop_len = len(stop)
            # make sure the result is not a stop word. e.g. foo should match foo, foo-123 but not foobar
            if result.startswith(stop) and (len(result) == stop_len or not str.isalpha(result[stop_len])):
                return parsy.Result.failure(index, "A-Za-z0-9_-:")

        if index <= len(stream) and valid_string_chars and found_no_number:
            return parsy.Result.success(index, result)
        else:
            return parsy.Result.failure(index, "A-Za-z0-9_-:")

    return unquoted_string_direct_parser


string_esc_dp = backslash_dp >> (backslash_dp | string("/") | string('"') | string("'"))
unquoted_string_dp = unquoted_string_parser()

single_quote_dp = string("'")
single_quoted_string_part_dp = regex(r"[^'\\]+")
single_quoted_string_part_or_esc_dp = (single_quoted_string_part_dp | string_esc_dp | backslash_dp).many().concat()
single_quoted_string_dp = single_quote_dp >> single_quoted_string_part_or_esc_dp << single_quote_dp

double_quote_dp = string('"')
double_quoted_string_part_dp = regex(r'[^"\\]+')
double_quoted_string_part_or_esc_dp = (double_quoted_string_part_dp | string_esc_dp | backslash_dp).many().concat()
double_quoted_string_dp = double_quote_dp >> double_quoted_string_part_or_esc_dp << double_quote_dp

any_string = parsy.any_char.many().concat()
any_non_white_space_string = parsy.test_char(lambda x: x != " ", "non whitespace").many().concat()
double_quoted_or_simple_string_dp = double_quoted_string_dp | any_string
any_non_whitespace_string = parsy.test_char(lambda c: c != " ", "non whitespace").at_least(1).concat()

true_dp = string("true").result(True)
false_dp = string("false").result(False)
null_dp = string("null").result(None)

# endregion

# region whitespace agnostic parser

lparen_p = lexeme(lparen_dp)
rparen_p = lexeme(rparen_dp)
l_bracket_p = lexeme(l_bracket_dp)
r_bracket_p = lexeme(r_bracket_dp)
l_curly_p = lexeme(l_curly_dp)
r_curly_p = lexeme(r_curly_dp)
gt_p = lexeme(gt_dp)
lt_p = lexeme(lt_dp)
colon_p = lexeme(colon_dp)
semicolon_p = lexeme(semicolon_dp)
comma_p = lexeme(comma_dp)
pipe_p = lexeme(pipe_dp)
dot_p = lexeme(dot_dp)
dot_dot_p = lexeme(dot_dot_dp)
equals_p = lexeme(equals_dp)
true_p = lexeme(true_dp)
false_p = lexeme(false_dp)
null_p = lexeme(null_dp)
float_p = lexeme(float_dp)
integer_p = lexeme(integer_dp)
variable_p = lexeme(variable_dp)
variable_no_array_p = lexeme(variable_no_array_dp)
literal_p = lexeme(literal_dp)
quoted_string_p = lexeme(double_quoted_string_dp)
unquoted_string_p = lexeme(unquoted_string_dp)

# endregion


@make_parser
def json_array_parser() -> Parser:
    yield l_bracket_p
    elements = yield json_value_p.sep_by(comma_p)
    yield r_bracket_p
    return elements


@make_parser
def json_object_pair() -> Parser:
    key = yield quoted_string_p | literal_p
    yield colon_p
    val = yield json_value_p
    return key, val


json_object_p = l_curly_dp >> json_object_pair.sep_by(comma_dp).map(dict) << r_curly_dp
simple_json_value_dp = (
    double_quoted_string_dp | true_dp | false_dp | null_dp | unquoted_string_dp | float_dp | integer_dp
)
json_value_dp = json_array_parser | json_object_p | simple_json_value_dp
json_value_p = lexeme(json_value_dp)



from copy import deepcopy
from distutils.command.config import config
from functools import lru_cache
import pytest
from attrs import define, field
from typing import ClassVar, Dict, Any
from resotolib.config import Config, ConfigNotFoundError
from resotolib.args import get_arg_parser, ArgumentParser
from resotolib.core import add_args as core_add_args


def test_config():
    arg_parser = get_arg_parser()
    core_add_args(arg_parser)
    arg_parser.parse_known_args()
    cfg = Config("test")
    cfg.add_config(ConfigTest)
    cfg.init_default_config()
    assert Config.dict() == {
        "configtest": {
            "testvar1": "testing123",
            "testvar2": 12345,
            "testvar3": {
                "mydict": {"foo": "bar", "abc": {"def": "ghi"}},
                "myint": 0,
                "mystr": "Hello",
            },
        }
    }
    cfg2 = Config("test2")
    assert cfg.configtest.testvar1 == cfg2.configtest.testvar1
    assert cfg.configtest.testvar1 == Config.configtest.testvar1
    assert cfg.configtest.testvar2 == cfg2.configtest.testvar2
    assert cfg.configtest.testvar3 == cfg2.configtest.testvar3
    Config.configtest.testvar2 += 1
    assert cfg.configtest.testvar2 == 12346
    with pytest.raises(ConfigNotFoundError):
        Config.does_not_exist.foo = "bar"
    with pytest.raises(ConfigNotFoundError):
        cfg.does_not_exist.foo = "bar"


def test_config_drop_deleted_entries():
    arg_parser = get_arg_parser()
    core_add_args(arg_parser)
    arg_parser.parse_known_args()
    cfg = Config("test")
    cfg.add_config(ConfigTest)

    config_from_the_core = {
        "configtest": {
            "testvar1": "1",
            "testvar2": 2,
            "testvar42": 42,
            "testvar3": {
                "myint": 42,
            },
        }
    }

    assert cfg.with_default_config(config_from_the_core) == {
        "configtest": {
            "testvar1": "1",
            "testvar2": 2,
            "testvar3": {
                "mydict": {"foo": "bar", "abc": {"def": "ghi"}},
                "myint": 42,
                "mystr": "Hello",
            },
        }
    }


def test_config_lru_cache():
    @lru_cache()
    def test_lru_config(config: Config):
        return config.configtest.testvar1

    arg_parser = get_arg_parser()
    core_add_args(arg_parser)
    arg_parser.parse_known_args()
    cfg = Config("test")
    cfg.add_config(ConfigTest)
    cfg.init_default_config()
    assert test_lru_config(cfg) == "testing123"
    # update config
    cfg.running_config.data["configtest"].testvar1 = "foo"
    cfg.running_config.revision = "foo_revision"
    assert cfg.configtest.testvar1 == "foo"
    assert test_lru_config(cfg) == "foo"
    # cleanup
    cfg.running_config.data["configtest"].testvar1 = "testing123"


def test_config_override():
    arg_parser = get_arg_parser()
    core_add_args(arg_parser)
    arg_parser.parse_known_args()
    cfg = Config("test")
    cfg.add_config(ConfigTest)
    cfg.init_default_config()
    assert Config.dict() == {
        "configtest": {
            "testvar1": "testing123",
            "testvar2": 12346,
            "testvar3": {
                "mydict": {"foo": "bar", "abc": {"def": "ghi"}},
                "myint": 0,
                "mystr": "Hello",
            },
        }
    }
    ArgumentParser.args.config_override = [
        "configtest.testvar1=testing124",
        "configtest.testvar3.myint=1",
        "configtest.testvar3.mystr=World",
        "configtest.testvar3.mydict.foo=baz",
        "configtest.testvar3.mydict.abc.def=jkl",
    ]
    cfg.override_config(cfg.running_config)
    assert Config.dict() == {
        "configtest": {
            "testvar1": "testing124",
            "testvar2": 12346,
            "testvar3": {
                "mydict": {"foo": "baz", "abc": {"def": "jkl"}},
                "myint": 1,
                "mystr": "World",
            },
        }
    }


@define
class NestedConfigTest:
    kind: ClassVar[str] = "nested_config_test"
    myint: int = field(default=0, metadata={"description": "My Int"})
    mystr: str = field(default="Hello", metadata={"description": "My String"})
    mydict: Dict[str, Any] = field(factory=lambda: {"foo": "bar", "abc": {"def": "ghi"}})


@define
class ConfigTest:
    kind: ClassVar[str] = "configtest"
    testvar1: str = field(default="testing123", metadata={"description": "A test string"})
    testvar2: int = field(default=12345, metadata={"description": "A test integer"})
    testvar3: NestedConfigTest = field(
        factory=lambda: NestedConfigTest(),
        metadata={"description": "A test of nested config"},
    )



import pytest
import sys
import tempfile
from contextlib import contextmanager
from resotolib.asynchronous.utils import async_lines, stdin_generator


class SimpleAsyncIterator:
    def __init__(self, data):
        self.data = data
        self.index = 0

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.index < len(self.data):
            value = self.data[self.index]
            self.index += 1
            return value
        else:
            raise StopAsyncIteration


@pytest.mark.parametrize(
    "input_data, expected_output",
    [
        ("line1\nline2\nline3", ["line1", "line2", "line3"]),
        ("line1\r\nline2\r\nline3", ["line1", "line2", "line3"]),
        ("line1\nline2\nline3\n", ["line1", "line2", "line3"]),
        ("line1\r\nline2\r\nline3\r\n", ["line1", "line2", "line3"]),
        ("", []),
    ],
)
@pytest.mark.asyncio
async def test_async_lines(input_data, expected_output):
    iterator = SimpleAsyncIterator(input_data)
    output = [line async for line in async_lines(iterator)]
    assert output == expected_output


@contextmanager
def replace_stdin(input_data: str):
    with tempfile.TemporaryFile(mode="w+t") as temp_file:
        temp_file.write(input_data)
        temp_file.seek(0)
        original_stdin = sys.stdin
        sys.stdin = temp_file
        try:
            yield
        finally:
            sys.stdin = original_stdin


@pytest.mark.parametrize(
    "input_data, expected_output",
    [
        ("line1\nline2\nline3", ["line1", "line2", "line3"]),
        ("line1\r\nline2\r\nline3", ["line1", "line2", "line3"]),
        ("", []),
    ],
)
@pytest.mark.asyncio
async def test_async_stdin_generator(input_data, expected_output):
    with replace_stdin(input_data):
        output = [line async for line in stdin_generator()]
    assert output == expected_output



from typing import Callable, Awaitable
from aiohttp.web import Request, StreamResponse


RequestHandler = Callable[[Request], Awaitable[StreamResponse]]
Middleware = Callable[[Request, RequestHandler], Awaitable[StreamResponse]]



import pytest
from parsy import ParseError

from resotolib.parse_util import (
    integer_dp,
    float_dp,
    json_value_p,
    unquoted_string_dp,
    unquoted_string_p,
    variable_p,
    unquoted_string_parser,
)


def test_variable_p() -> None:
    assert variable_p.parse("foo.bla.bar") == "foo.bla.bar"
    assert variable_p.parse("foo.bla.bar[*]") == "foo.bla.bar[*]"
    assert variable_p.parse("foo.bla.`foo-bar`") == "foo.bla.`foo-bar`"
    assert variable_p.parse("foo.bla.bar[*].`:-)`.bla[1].test") == "foo.bla.bar[*].`:-)`.bla[1].test"


def test_number_parse() -> None:
    assert integer_dp.parse("-123") == -123
    assert integer_dp.parse("123") == 123
    assert integer_dp.parse("+123") == 123
    assert float_dp.parse("-123.321") == -123.321
    assert float_dp.parse("123.321") == 123.321
    assert float_dp.parse("+123.321") == 123.321


def test_unquoted_string_dp() -> None:
    assert unquoted_string_p.parse("a") == "a"
    assert unquoted_string_p.parse("  abc:123  ") == "abc:123"
    assert unquoted_string_dp.parse("2021-09-23T06:42:26Z") == "2021-09-23T06:42:26Z"
    with pytest.raises(ParseError) as ex:
        unquoted_string_dp.parse("2021%")
    assert str(ex.value) == "expected 'A-Za-z0-9_-:' at 0:5"
    # not will not be parsed, but words
    p = unquoted_string_parser("not", "test", "foo")
    assert p.parse("nothing") == "nothing"
    assert p.parse("tester") == "tester"
    assert p.parse("fooo") == "fooo"
    assert p.parse("no") == "no"
    with pytest.raises(ParseError):
        assert p.parse("not")  # stop word
    with pytest.raises(ParseError):
        assert p.parse("not_123")  # stop word
    with pytest.raises(ParseError):
        assert p.parse("test")  # stop word


def test_json_value_p() -> None:
    assert json_value_p.parse("-123") == -123
    assert json_value_p.parse("123") == 123
    assert json_value_p.parse("123.23") == 123.23
    assert json_value_p.parse("-123.23") == -123.23
    assert json_value_p.parse("127.0.0.1") == "127.0.0.1"
    assert json_value_p.parse('"-123"') == "-123"
    assert json_value_p.parse("2021-09-23T06:42:26Z") == "2021-09-23T06:42:26Z"
    assert json_value_p.parse('"2021-09-23T06:42:26Z"') == "2021-09-23T06:42:26Z"
    assert json_value_p.parse("true") is True
    assert json_value_p.parse("false") is False
    assert json_value_p.parse("null") is None
    assert json_value_p.parse('["a", 1, "2", {"test":"a"}]') == ["a", 1, "2", {"test": "a"}]
    assert json_value_p.parse('{"test":{"foo":{"bla":123}}}') == {"test": {"foo": {"bla": 123}}}
    assert json_value_p.parse("test") == "test"
    assert json_value_p.parse("[foo, bla, bar,baz]") == ["foo", "bla", "bar", "baz"]



from __future__ import annotations

import asyncio
import logging
from asyncio import Task, AbstractEventLoop
from contextlib import suppress
from datetime import timedelta
from typing import Any, Callable, Optional, Awaitable

log = logging.getLogger(__name__)


class Periodic:
    """
    Periodic execution of a function based on a defined frequency that can be started and stopped.
    """

    def __init__(
        self,
        name: str,
        func: Callable[[], Any],
        frequency: timedelta,
        first_run: Optional[timedelta] = None,
        loop: Optional[AbstractEventLoop] = None,
    ):
        self.name = name
        self.func = func
        self.frequency = frequency
        self.first_run = first_run if first_run else frequency
        self._task: Optional[Task[None]] = None
        self._loop = loop

    @property
    def started(self) -> bool:
        return self._task is not None

    async def start(self) -> None:
        if self._task is None:
            # Start task to call func periodically:
            self._task = asyncio.ensure_future(self._run(), loop=self._loop)
            log.debug(f"Periodic task {self.name} has been started.")

    async def stop(self) -> None:
        # Stop task and await it stopped:
        if self._task is not None:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    async def _run(self) -> None:
        await asyncio.sleep(self.first_run.total_seconds())
        while True:
            log.debug(f"Execute periodic task {self.name}.")
            try:
                result = self.func()
                if isinstance(result, Awaitable):
                    await result
            except Exception as ex:
                log.error(
                    f"Periodic function {self.name} caught an exception: {ex}",
                    exc_info=ex,
                )
            await asyncio.sleep(self.frequency.total_seconds())



import asyncio
from datetime import timedelta
from typing import Any

import pytest
from aiohttp.test_utils import TestClient
from multidict import CIMultiDict
from pytest import mark
from aiohttp.web import Application, Request, Response
from resotolib.asynchronous.web.auth import check_auth, jwt_from_context
from resotolib.jwt import encode_jwt

# noinspection PyUnresolvedReferences
from aiohttp.pytest_plugin import aiohttp_client


@pytest.fixture
async def loop() -> Any:
    return asyncio.get_running_loop()


@pytest.fixture
async def app_with_auth() -> Application:
    async def hello(_: Request) -> Response:
        jwt = await jwt_from_context()
        # make sure, the context variable is set
        assert jwt["foo"] == "bla"
        assert "exp" in jwt
        return Response(text="Hello, world")

    app = Application(middlewares=[check_auth("test", timedelta(minutes=5), set())])
    app.router.add_get("/", hello)
    return app


@mark.asyncio
async def test_correct_psk(aiohttp_client: Any, app_with_auth: Application) -> None:
    client: TestClient = await aiohttp_client(app_with_auth)
    jwt = encode_jwt({"foo": "bla"}, "test")
    resp = await client.get("/", headers=CIMultiDict({"Authorization": f"Bearer {jwt}"}))
    assert resp.status == 200


@mark.asyncio
async def test_wrong_psk(aiohttp_client: Any, app_with_auth: Application) -> None:
    client: TestClient = await aiohttp_client(app_with_auth)
    jwt = encode_jwt({"foo": "bla"}, "wrong!")
    resp = await client.get("/", headers=CIMultiDict({"Authorization": f"Bearer {jwt}"}))
    assert resp.status == 401


@mark.asyncio
async def test_no_psk(aiohttp_client: Any, app_with_auth: Application) -> None:
    client: TestClient = await aiohttp_client(app_with_auth)
    resp = await client.get("/")
    assert resp.status == 401



import os
import tempfile
import time

import cherrypy
import requests

from resotolib.utils import get_free_port
from resotolib.web import WebServer
from resotolib.web.metrics import WebApp
from resotolib.x509 import (
    gen_rsa_key,
    gen_csr,
    bootstrap_ca,
    sign_csr,
    write_cert_to_file,
    write_key_to_file,
)


def test_web():
    # Find a free local port to reuse when we bind the web server.
    # This is so that multiple builds/tests can run in parallel
    # on the same CI agent.
    # todo: race between closing socket and reusing free port in WebServer
    free_port = get_free_port()
    print(f"Starting http webserver on port {free_port}")
    web_server = WebServer(WebApp(), web_port=free_port)
    web_server.daemon = True
    web_server.start()
    start_time = time.time()
    while not web_server.serving:
        if time.time() - start_time > 10:
            raise RuntimeError("timeout waiting for web server start")
        time.sleep(0.1)

    # We're statically using localhost in the endpoint url.
    # Other options would have been to set ArgumentParser.args.web_host
    # and then connect to that value. However we'd have to use an IP
    # address and then needed to decide if we use either
    # 127.0.0.1 or ::1. Which might fail on CI boxes without
    # IPv4 or IPv6 respectively. Instead we leave the default which
    # binds to all IPs and assume that localhost will resolve to
    # the appropriate v4 or v6 loopback address. A disadvantage
    # of this is that for a brief moment during the test we're
    # exposing the web server on all local IPs.
    endpoint = f"http://localhost:{free_port}"
    r = requests.get(f"{endpoint}/health")
    assert r.text == "ok\r\n"
    web_server.shutdown()
    while web_server.is_alive():
        print("Waiting for web server to shutdown")
        time.sleep(1)


def disabled_test_secure_web():
    with tempfile.TemporaryDirectory() as tmp:
        ca_key, ca_cert = bootstrap_ca()
        cert_key = gen_rsa_key()
        cert_csr = gen_csr(cert_key, common_name="localhost")
        cert_crt = sign_csr(cert_csr, ca_key, ca_cert)
        ca_cert_path = os.path.join(tmp, "ca.crt")
        cert_key_path = os.path.join(tmp, "cert.key")
        cert_crt_path = os.path.join(tmp, "cert.crt")

        write_cert_to_file(ca_cert, cert_path=ca_cert_path)
        write_key_to_file(cert_key, key_path=cert_key_path)
        write_cert_to_file(cert_crt, cert_path=cert_crt_path)

        free_port = get_free_port()
        print(f"Starting https webserver on port {free_port}")
        web_server = WebServer(WebApp(), web_port=free_port, ssl_cert=cert_crt_path, ssl_key=cert_key_path)
        web_server.daemon = True
        web_server.start()
        start_time = time.time()
        while not web_server.serving:
            if time.time() - start_time > 10:
                raise RuntimeError("timeout waiting for web server start")
            time.sleep(0.1)

        endpoint = f"https://localhost:{free_port}"
        r = requests.get(f"{endpoint}/health", verify=ca_cert_path)
        assert r.text == "ok\r\n"
        web_server.shutdown()
        while web_server.is_alive():
            print("Waiting for web server to shutdown")
            time.sleep(1)



import os
from abc import ABC
from attrs import define
from typing import ClassVar

import pytest
from pytest import raises

from resotolib.baseresources import BaseResource
from resotolib.core.model_check import check_overlap


@define(slots=False)
class BreakingResource(BaseResource, ABC):
    kind: ClassVar[str] = "breaking"
    volume_size: str = ""


@pytest.mark.skipif(os.environ.get("MODEL_CHECK") is None, reason="Model check is disabled")
def test_check() -> None:
    # this will throw an exception, since breaking resource has a breaking property
    with raises(Exception):
        check_overlap()
    # hacky way to "delete" the fields - the exporter will not see the field any longer.
    BreakingResource.__attrs_attrs__ = {}
    check_overlap()



import os
import resotolib
import pkg_resources
from setuptools import setup, find_packages


def read(file_name: str) -> str:
    with open(os.path.join(os.path.dirname(__file__), file_name)) as of:
        return of.read()


setup(
    name=resotolib.__title__,
    version=resotolib.__version__,
    description=resotolib.__description__,
    license=resotolib.__license__,
    packages=find_packages(),
    package_data={"resotolib": ["py.typed"]},
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    include_package_data=True,
    zip_safe=False,
    install_requires=[str(requirement) for requirement in pkg_resources.parse_requirements(read("requirements.txt"))],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    classifiers=[
        # Current project status
        "Development Status :: 4 - Beta",
        # Audience
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
        # License information
        "License :: OSI Approved :: Apache Software License",
        # Supported python versions
        "Programming Language :: Python :: 3.9",
        # Supported OS's
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        # Extra metadata
        "Environment :: Console",
        "Natural Language :: English",
        "Topic :: Security",
        "Topic :: Utilities",
    ],
    keywords="cloud security",
    url="https://github.com/someengineering/resoto/tree/main/resotolib",
)



import json
from datetime import timedelta
from typing import TypeVar, Any, Type, Optional, Dict, Union, List

import attrs
import cattrs
import jsons
from cattrs import override
from cattrs.gen import make_dict_unstructure_fn
from jsons import set_deserializer

from resotolib.durations import parse_duration
from resotolib.logger import log
from resotolib.types import Json, JsonElement

AnyT = TypeVar("AnyT")


converter = cattrs.Converter()
converter.register_unstructure_hook_factory(
    attrs.has,
    lambda cls: make_dict_unstructure_fn(
        cls, converter, **{a.name: override(omit=True) for a in attrs.fields(cls) if a.name.startswith("_")}
    ),
)


def to_json(node: Any, **kwargs: Any) -> Json:
    """
    Use this method, if the given node is known as complex object,
    so the result will be a json object.
    """
    unstructured = converter.unstructure(node)
    if strip_attr := kwargs.get("strip_attr"):
        if isinstance(strip_attr, str):
            if unstructured.get(strip_attr):
                del unstructured[strip_attr]
        elif isinstance(strip_attr, (list, tuple)):
            for field in strip_attr:
                if unstructured.get(field):
                    del unstructured[field]
        else:
            raise ValueError(f"Don't know how to handle type of strip_attr: {strip_attr}")
    result: Json = jsons.dump(
        unstructured,
        strip_microseconds=True,
        strip_nulls=True,
        # class variables have to stripped manually via strip_attr=
        # see: https://github.com/ramonhagenaars/jsons/issues/177
        # strip_class_variables=True,
        **kwargs,
    )
    return result


def to_json_str(node: Any, json_kwargs: Optional[Dict[str, object]] = None, **kwargs: Any) -> str:
    """
    Json string representation of the given object.
    """
    return json.dumps(to_json(node, **kwargs), **(json_kwargs or {}))  # type: ignore


def from_json(json: JsonElement, clazz: Type[AnyT], **kwargs: Any) -> AnyT:
    """
    Loads a json object into a python object.
    :param json: the json object to load.
    :param clazz: the type of the python object.
    :return: the loaded python object.
    """
    try:
        return jsons.load(json, clazz, **kwargs) if clazz != dict else json  # type: ignore
    except Exception as e:
        log.debug(f"Can not deserialize json into class {clazz.__name__}: {json}. Error: {e}")
        raise


def value_in_path(element: JsonElement, path_or_name: Union[List[str], str]) -> Optional[Any]:
    """
    Access a value in a json object by a defined path.
    {"a": {"b": {"c": 1}}} -> value_in_path({"a": {"b": {"c": 1}}}, ["a", "b", "c"]) -> 1
    The path can be defined as a list of strings or as a string with dots as separator.
    """
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    at = len(path)

    def at_idx(current: JsonElement, idx: int) -> Optional[Any]:
        if at == idx:
            return current
        elif current is None or not isinstance(current, dict) or path[idx] not in current:
            return None
        else:
            return at_idx(current[path[idx]], idx + 1)

    return at_idx(element, 0)


def set_value_in_path(element: JsonElement, path_or_name: Union[List[str], str], js: Optional[Json] = None) -> Json:
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    at = len(path) - 1

    def at_idx(current: Json, idx: int) -> None:
        if at == idx:
            current[path[-1]] = element
        else:
            value = current.get(path[idx])
            if not isinstance(value, dict):
                value = {}
                current[path[idx]] = value
            at_idx(value, idx + 1)

    js = js if js is not None else {}
    at_idx(js, 0)
    return js


# allow timedelta either as number of seconds or as duration string
def timedelta_from_json(js: Any, _: type = object, **__: Any) -> timedelta:
    if isinstance(js, str):
        return parse_duration(js)
    elif isinstance(js, (int, float)):
        return timedelta(seconds=js)
    else:
        raise ValueError(f"Cannot convert {js} to timedelta")


def is_empty(js: JsonElement) -> bool:
    if js is None:
        return True
    elif isinstance(js, dict):
        return all(is_empty(v) for v in js.values())
    elif isinstance(js, list):
        return all(is_empty(v) for v in js)
    else:
        return False


set_deserializer(timedelta_from_json, timedelta)



from textwrap import dedent
from typing import Callable, Any, List

import yaml
from pytest import fixture, raises

from resotolib.core.progress import ProgressDone, Progress, ProgressInfo, ProgressTree


@fixture
def progress() -> ProgressTree:
    fp = Progress.from_progresses
    result = fp(
        "aws",
        [
            ProgressDone("region1", 1, 2, path=["account1"]),
            ProgressDone("region2", 2, 2, path=["account1"]),
            ProgressDone("region1", 1, 2, path=["account2"]),
            ProgressDone("region3", 2, 2, path=["account2"]),
            ProgressDone("region2", 1, 2, path=["account3"]),
            ProgressDone("region3", 2, 2, path=["account3"]),
            ProgressDone("extra", 0, 100, path=["some", "deeply", "nested"]),
        ],
    )
    return result


def test_tree() -> None:
    a = ProgressDone("a", 0, 1, path=["cloud", "account", "region"])
    b = ProgressDone("b", 0, 1, path=["cloud", "account", "region"])
    p = Progress.from_progresses("123", [a, b])
    assert len(p.sub_tree.children("root.cloud.account.region")) == 2


def test_info_json() -> None:
    a = ProgressDone("a", 0, 1, path=["cloud", "account", "region"])
    b = ProgressDone("b", 16, 23, path=["cloud", "account", "region"])
    c = ProgressDone("c", 12, 12, path=["cloud", "account", "region"])
    p = Progress.from_progresses("123", [a, b, c])
    assert p.info_json() == {"123": {"cloud": {"account": {"region": {"a": "in progress", "b": "69%", "c": "done"}}}}}


def test_sub_progress(progress: ProgressTree) -> None:
    sub = progress.sub_progress("root.some.deeply")
    assert sub.overall_progress() == ProgressInfo(0, 100)
    assert sub.percentage == 0


def test_progress(progress: Progress) -> None:
    assert progress.overall_progress() == ProgressInfo(225, 400)
    assert progress.percentage == 56


def test_marshalling(progress: Progress) -> None:
    assert Progress.from_json(progress.to_json()).to_json() == progress.to_json()


def test_invalid() -> None:
    with raises(ValueError):
        ProgressDone("region1", 1, 0)


def test_path(progress: ProgressTree) -> None:
    assert progress.has_path("account1.region1")
    res = progress.by_path("account1.region1")
    assert res is not None
    assert res.percentage == 50

    sub = progress.sub_progress("account1")
    assert sub is not None
    assert sub.percentage == 75


def test_equality(progress: ProgressTree) -> None:
    pgc = progress.copy()
    assert pgc == progress
    pgc.add_progress(ProgressDone("region1", 2, 2, path=["account1"]))
    assert pgc != progress


def test_order() -> None:
    def order(key: Callable[[Progress], Any]) -> List[str]:
        pt = ProgressTree("test")
        pt.add_progress(ProgressDone("a", 10, 10))
        pt.add_progress(ProgressDone("b", 8, 10))
        pt.add_progress(ProgressDone("c", 3, 10))
        pt.add_progress(ProgressDone("d", 1, 10))
        js = pt.to_json(key=key)
        return [x["name"] for x in js["parts"]]

    assert order(lambda x: x.name) == ["a", "b", "c", "d"]
    assert order(lambda x: x.overall_progress().percentage) == ["d", "c", "b", "a"]


def test_nested_progress() -> None:
    accounts = ["a1", "a2", "a3"]
    regions = ["r1", "r2", "r3"]

    overall_progress = ProgressTree("test")

    regional_progress = ProgressTree("aws")
    for account in accounts:
        regional_progress.add_progress(ProgressDone(account, 0, 1))
    overall_progress.add_progress(regional_progress)
    assert (
        yaml.safe_dump(overall_progress.info_json(), indent=2)
        == dedent(
            """
            test:
              aws:
                a1: in progress
                a2: in progress
                a3: in progress
            """
        ).lstrip()
    )

    results = []
    for account in accounts:
        regional_progress = ProgressTree(account, path=["aws"])
        for region in regions:
            regional_progress.add_progress(ProgressDone(region, 0, 1))
        overall_progress.add_progress(regional_progress)
        results.append(overall_progress.info_json())
    assert (
        yaml.safe_dump(results, indent=2)
        == dedent(
            """
            - test:
                aws:
                  a1:
                    r1: in progress
                    r2: in progress
                    r3: in progress
                  a2: in progress
                  a3: in progress
            - test:
                aws:
                  a1:
                    r1: in progress
                    r2: in progress
                    r3: in progress
                  a2:
                    r1: in progress
                    r2: in progress
                    r3: in progress
                  a3: in progress
            - test:
                aws:
                  a1:
                    r1: in progress
                    r2: in progress
                    r3: in progress
                  a2:
                    r1: in progress
                    r2: in progress
                    r3: in progress
                  a3:
                    r1: in progress
                    r2: in progress
                    r3: in progress
            """
        ).lstrip()
    )



from datetime import timedelta
from itertools import chain
from typing import Callable, Any, cast, Optional, TypeVar

from hypothesis import given
from hypothesis.strategies import sampled_from, tuples, integers, composite, lists, SearchStrategy, just

from resotolib.durations import time_unit_parser, time_units, parse_duration, DurationRe, duration_str

units_gen = sampled_from(list(chain.from_iterable(names for _, names, _ in time_units)))
combines_gen = sampled_from(["", ", ", " and "])
duration_gen = tuples(integers(1000, 1000), units_gen).map(lambda x: f"{x[0]}{x[1]}")

T = TypeVar("T")
UD = Callable[[SearchStrategy[Any]], Any]


def optional(st: SearchStrategy[T]) -> SearchStrategy[Optional[T]]:
    return st | just(None)


class Drawer:
    """
    Only here for getting a drawer for typed drawings.
    """

    def __init__(self, hypo_drawer: Callable[[SearchStrategy[Any]], Any]):
        self._drawer = hypo_drawer

    def draw(self, st: SearchStrategy[T]) -> T:
        return cast(T, self._drawer(st))

    def optional(self, st: SearchStrategy[T]) -> Optional[T]:
        return self.draw(optional(st))


@composite
def durations_gen(ud: UD) -> str:
    d = Drawer(ud)
    result = d.draw(sampled_from(["", "+", "-"]))
    first = True
    for duration in d.draw(lists(duration_gen, min_size=1, max_size=4)):
        if first:
            first = False
        else:
            result += d.draw(combines_gen)
        result += duration
    return result


@given(durations_gen())
def test_arbitrary_durations(duration_str: str) -> None:
    assert DurationRe.fullmatch(duration_str)
    parse_duration(duration_str)


def test_parse_duration() -> None:
    for short, names, seconds in time_units:
        for name in names:
            assert time_unit_parser.parse(name) == seconds

    assert parse_duration("1s") == timedelta(seconds=1)
    assert parse_duration("4d") == timedelta(days=4)
    assert parse_duration("1h") == timedelta(hours=1)
    assert parse_duration("32days, 4hours and 3min and 3s") == timedelta(days=32, hours=4, minutes=3, seconds=3)
    assert parse_duration("-32days, 4hours and 3min and 3s") == timedelta(days=-32, hours=-4, minutes=-3, seconds=-3)
    assert parse_duration("3d4h6m5s") == timedelta(days=3, hours=4, minutes=6, seconds=5)


def test_duration_string() -> None:
    duration = timedelta(days=1, hours=2, minutes=3, seconds=4)
    assert duration_str(duration) == "1d2h3min4s"
    # define the most granular unit
    assert duration_str(duration, down_to_unit="min") == "1d2h3min"
    assert duration_str(duration, down_to_unit="h") == "1d2h"
    assert duration_str(duration, down_to_unit="d") == "1d"
    assert duration_str(duration, down_to_unit="min") == "1d2h3min"
    # define the precision
    assert duration_str(duration, precision=1) == "1d"
    assert duration_str(duration, precision=2) == "1d2h"
    assert duration_str(duration, precision=3) == "1d2h3min"



from attrs import define, field
from datetime import datetime
from enum import Enum
from typing import List, Optional, Dict, Union, ClassVar, Literal

from resotolib.core.model_export import (
    is_collection,
    type_arg,
    is_dict,
    dict_types,
    transitive_classes,
    dataclasses_to_resotocore_model,
    model_name,
    dynamic_object_to_resotocore_model,
    is_primitive_or_primitive_union,
)
from resotolib.baseresources import ModelReference


class ExampleEnum(Enum):
    Moon = "moon"
    Sun = "sun"
    Earth = "earth"


@define(slots=False)
class DataClassBase:
    kind: ClassVar[str] = "base"
    successor_kinds: ClassVar[Dict[str, List[str]]] = {"default": [], "delete": []}

    tags: Dict[str, str] = field(metadata={"description": "Description of tags"})
    _private_prop: dict
    __dunder_prop: list
    ctime: Optional[datetime] = field(metadata={"synthetic": {"age": "trafo.duration_to_datetime"}})


@define
class DataClassProp:
    kind: ClassVar[str] = "prop"
    key: Optional[str]
    value: Union[str, int, float]


@define(slots=False)
class DataClassExample(DataClassBase):
    kind: ClassVar[str] = "example"
    successor_kinds: ClassVar[Dict[str, List[str]]] = {
        "default": ["example"],
    }
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["base"]},
        "predecessors": {"delete": ["other"]},
    }
    list_of_string: List[str]
    optional_list_of_props: Optional[List[DataClassProp]]
    props: List[DataClassProp]
    values: dict
    simple_list: list
    weird_list: List[Dict[str, Dict[str, Dict[str, Dict[str, DataClassProp]]]]]
    weird_dict: Dict[str, Dict[str, Dict[str, Dict[str, DataClassProp]]]]
    optionally_weird_dict: Optional[Optional[Dict[str, Dict[str, Dict[str, Dict[Union[str, int], DataClassProp]]]]]]
    some_enum: ExampleEnum
    optional_enum: Optional[ExampleEnum]


@define(slots=False)
class DataClassOther(DataClassBase):
    kind: ClassVar[str] = "other"
    something: str


def test_collection() -> None:
    assert is_collection(Optional[List[str]]) is True
    assert is_collection(List[str]) is True
    assert is_collection(list) is True
    assert is_collection(dict) is False

    assert type_arg(Optional[List[int]]) == int
    assert type_arg(List[datetime]) == datetime


def test_dictionary() -> None:
    assert is_dict(Optional[Dict[str, str]]) is True
    assert is_dict(Dict[str, str]) is True
    assert is_dict(dict) is True
    assert is_dict(list) is False

    assert dict_types(Dict[str, int]) == (str, int)
    assert dict_types(dict) == (object, object)


def test_transitive() -> None:
    assert transitive_classes({DataClassExample}) == {
        DataClassExample,
        DataClassProp,
        DataClassBase,
        DataClassOther,
        ExampleEnum,
    }


def test_model_name() -> None:
    assert model_name(str) == "string"
    assert model_name(int) == "int64"
    assert model_name(Dict[str, str]) == "dictionary[string, string]"
    assert model_name(Optional[Dict[str, int]]) == "dictionary[string, int64]"
    assert model_name(list) == "any[]"
    assert model_name(dict) == "dictionary[any, any]"
    assert model_name(List[DataClassProp]) == "prop[]"
    assert model_name(Optional[List[DataClassProp]]) == "prop[]"
    assert (
        model_name(Optional[Dict[str, List[Dict[str, DataClassProp]]]])
        == "dictionary[string, dictionary[string, prop][]]"
    )


def test_enum_to_model() -> None:
    assert dataclasses_to_resotocore_model({ExampleEnum}) == [
        {
            "fqn": "example_enum",
            "runtime_kind": "string",
            "enum": ["moon", "sun", "earth"],
        }
    ]


def test_dataclasses_to_resotocore_model() -> None:
    result = dataclasses_to_resotocore_model({DataClassExample})
    assert len(result) == 5
    for r in result:
        props = {p["name"]: p for p in r.get("properties", [])}
        if r["fqn"] == "base":
            assert len(r["properties"]) == 3
            assert r["successor_kinds"] == {"default": [], "delete": []}
            assert props["tags"]["kind"] == "dictionary[string, string]"
            assert props["tags"]["description"] == "Description of tags"
            assert props["ctime"]["kind"] == "datetime"
            assert props["age"]["kind"] == "trafo.duration_to_datetime"
            assert props["age"]["synthetic"]["path"] == ["ctime"]
        elif r["fqn"] == "prop":
            assert r["successor_kinds"] is None
            assert len(r["properties"]) == 2
            assert props["key"]["kind"] == "string"
            assert props["value"]["kind"] == "any"
        elif r["fqn"] == "other":
            assert r["successor_kinds"] == {"default": [], "delete": ["example"]}
        elif r["fqn"] == "example":
            assert set(r["successor_kinds"]["default"]) == {"base", "example"}
            assert len(r["properties"]) == 10
            assert props["list_of_string"]["kind"] == "string[]"
            assert props["optional_list_of_props"]["kind"] == "prop[]"
            assert props["props"]["kind"] == "prop[]"
            assert props["values"]["kind"] == "dictionary[any, any]"
            assert props["simple_list"]["kind"] == "any[]"
            assert (
                props["weird_list"]["kind"]
                == "dictionary[string, dictionary[string, dictionary[string, dictionary[string, prop]]]][]"
            )
            assert (
                props["weird_dict"]["kind"]
                == "dictionary[string, dictionary[string, dictionary[string, dictionary[string, prop]]]]"
            )
            assert (
                props["optionally_weird_dict"]["kind"]
                == "dictionary[string, dictionary[string, dictionary[string, dictionary[any, prop]]]]"
            )
            assert props["some_enum"]["kind"] == "example_enum"
            assert props["optional_enum"]["kind"] == "example_enum"
        elif r["fqn"] == "example_enum":
            assert r["runtime_kind"] == "string"
            assert r["enum"] == [e.value for e in ExampleEnum]


@define
class AwsTestConfig:
    kind: ClassVar[str] = "aws_config"
    access_key: str = field(metadata={"description": "The AWS access key."})
    secret_key: str = field(metadata={"description": "The secret part of the key."})


@define
class GcpTestConfigConfig:
    kind: ClassVar[str] = "gcp_config"
    foo: int = field(metadata={"description": "Some foo value."})


def test_config_export():
    # Let's assume a dynamic top level object of name Config
    # The properties are defined by name and related type.
    result = dynamic_object_to_resotocore_model("config", {"aws": AwsTestConfig, "gcp": GcpTestConfigConfig})
    result_dict = {a["fqn"]: a for a in result}
    assert len(result_dict["gcp_config"]["properties"]) == 1
    assert len(result_dict["aws_config"]["properties"]) == 2
    # Aws properties are rendered with description
    aws = {a["name"]: a["description"] for a in result_dict["aws_config"]["properties"]}
    assert aws == {
        "access_key": "The AWS access key.",
        "secret_key": "The secret part of the key.",
    }
    # Gcp properties are rendered with description
    gcp = {a["name"]: a["description"] for a in result_dict["gcp_config"]["properties"]}
    assert gcp == {"foo": "Some foo value."}
    # All global config properties are defined
    config = {a["name"] for a in result_dict["config"]["properties"]}
    assert config == {"aws", "gcp"}


def test_primitive_union() -> None:
    # simple types
    assert is_primitive_or_primitive_union(str) is True
    assert is_primitive_or_primitive_union(int) is True
    assert is_primitive_or_primitive_union(bool) is True
    assert is_primitive_or_primitive_union(int) is True
    assert is_primitive_or_primitive_union(float) is True
    assert is_primitive_or_primitive_union(type(None)) is True
    assert is_primitive_or_primitive_union(GcpTestConfigConfig) is False
    # literal types
    assert is_primitive_or_primitive_union(Literal["test"]) is True
    # union types
    assert is_primitive_or_primitive_union(Union[str, int, None]) is True
    assert is_primitive_or_primitive_union(Union[str, int, GcpTestConfigConfig]) is False
    assert is_primitive_or_primitive_union(Union[float]) is True
    assert is_primitive_or_primitive_union(Optional[str]) is True
    assert is_primitive_or_primitive_union(Optional[int]) is True
    assert is_primitive_or_primitive_union(Optional[GcpTestConfigConfig]) is False



import jsons
import threading
import os

from resotolib.json import from_json
from resotolib.logger import log
from resotolib.args import ArgumentParser, convert
from resotolib.core.ca import TLSData
from resotolib.proc import restart
from resotolib.core.model_export import dataclasses_to_resotocore_model, optional_origin
from resotolib.core import ResotocoreURI
from resotolib.core.config import (
    get_config,
    set_config,
    ConfigNotFoundError,
    update_config_model,
)
from resotolib.core.events import CoreEvents
from resotolib.utils import replace_env_vars, merge_json_elements, drop_deleted_attributes
from resotolib.types import Json
from typing import Dict, Any, List, Optional, Type, cast
from attrs import fields


class RunningConfig:
    def __init__(self) -> None:
        """Initialize the global config."""
        self.data: Dict[str, Any] = {}
        self.revision: str = ""
        self.classes: Dict[str, type] = {}
        self.types: Dict[str, Dict[str, type]] = {}

    def apply(self, other: "RunningConfig") -> None:
        """Apply another config to this one.

        Only updates references, does not create a copy of the data.
        """
        if isinstance(other, RunningConfig):
            self.data = other.data
            self.revision = other.revision
            self.classes = other.classes
            self.types = other.types
        else:
            raise TypeError(f"Cannot apply {type(other)} to RunningConfig")


_config = RunningConfig()


class MetaConfig(type):
    def __getattr__(cls, name: str) -> Any:
        if name in _config.data:
            return _config.data[name]
        else:
            raise ConfigNotFoundError(f"No such config {name}")


class Config(metaclass=MetaConfig):
    running_config: RunningConfig = _config

    def __init__(
        self,
        config_name: str,
        resotocore_uri: Optional[str] = None,
        tls_data: Optional[TLSData] = None,
    ) -> None:
        self._config_lock = threading.Lock()
        self.config_name = config_name
        self._initial_load = True
        resotocore = ResotocoreURI(resotocore_uri)
        self.resotocore_uri = resotocore.http_uri
        self.verify = None
        if tls_data:
            self.verify = tls_data.verify
        self._ce = CoreEvents(
            resotocore.ws_uri,
            events={"config-updated"},
            message_processor=self.on_config_event,
            tls_data=tls_data,
        )

    def __getattr__(self, name: str) -> Any:
        if name in self.running_config.data:
            return self.running_config.data[name]
        else:
            raise ConfigNotFoundError(f"No such config {name}")

    def shutdown(self) -> None:
        self._ce.shutdown()

    @staticmethod
    def init_default_config() -> None:
        for config_id, config_data in Config.running_config.classes.items():
            if config_id not in Config.running_config.data:
                log.debug(f"Initializing defaults for config section {config_id}")
                Config.running_config.data[config_id] = config_data()

    @staticmethod
    def add_config(config: object) -> None:
        """Add a config to the config manager.

        Takes a dataclass as input and adds its fields to the config store.
        Dataclass must have a kind ClassVar which specifies the top level config name.
        """
        if hasattr(config, "kind"):
            Config.running_config.classes[config.kind] = config  # type: ignore
            Config.running_config.types[config.kind] = {}
            for field in fields(config):  # type: ignore
                if hasattr(field, "type"):
                    Config.running_config.types[config.kind][field.name] = optional_origin(field.type)
        else:
            raise RuntimeError("Config must have a 'kind' attribute")

    def load_config(self, reload: bool = False) -> None:
        if len(Config.running_config.classes) == 0:
            raise RuntimeError("No config added")
        with self._config_lock:
            config_json = {}
            raw_config_json = {}
            new_config = None
            new_config_revision = None

            ############################################################
            # Load the config from the core, populate it with defaults
            # and send it back to resotocore
            ############################################################
            try:
                # We need two configs here: one with env_vars resolved and overrides applied
                # and one with the raw config as it was stored in the database
                config_response, new_config_revision = get_config(
                    self.config_name, self.resotocore_uri, verify=self.verify  # type: ignore
                )
                # config with env_vars resolved and overrides applied
                config_json = config_response["config"]
                config_json = {
                    k: replace_env_vars(v, os.environ, keep_unresolved=False) for k, v in config_json.items()
                }
                # raw config as it was stored in the database, to be sent to the core
                raw_config_json = config_response["raw_config"]

                if len(config_json) == 0:
                    if self._initial_load:
                        raise ConfigNotFoundError("Empty config returned - loading defaults")
                    else:
                        raise ValueError("Empty config returned")
            except ConfigNotFoundError:
                pass
            else:
                log.info(f"Loaded config {self.config_name} revision {new_config_revision}")
                new_config = Config.read_config(
                    config_json, reason="loading new config to check if restart is required"
                )
                # test if the config has changed using the most latest version
                # with overrides applied and env_vars resolved
                if reload and self.restart_required(new_config):
                    restart()

            # update the global config object with the confif from the core
            Config.running_config.data = self.with_default_config(raw_config_json)
            # if the raw_config was not empty, we need to set the revision too
            if new_config_revision:
                Config.running_config.revision = new_config_revision

            # now we're ready to send the raw config plus the new defaults to the core
            if self._initial_load:
                # Try to store the generated config. Handle failure gracefully.
                try:
                    # Send the raw_config + default values to resotocore
                    self.save_config()
                except RuntimeError as e:
                    log.error(f"Failed to save config: {e}")

            ############################################################
            # Once the core got the raw config, use the config with
            # overrides applied and env_vars resolved
            ############################################################
            # merge the config with overrides and env_vars into the default config
            self.init_default_config()
            default_config_dict = self.dict()

            config_with_defaults = cast(Json, merge_json_elements(default_config_dict, config_json))
            self.apply_path_overrides_resolve_env_vars(Config.running_config, config_with_defaults)
            self.override_config(Config.running_config)
            self._initial_load = False
            if not self._ce.is_alive():
                log.debug("Starting config event listener")
                self._ce.start()

    def with_default_config(self, raw_config_json: Json) -> Json:
        """
        Merges the raw config from the core with the default config and clean up deleted entries.
        """

        # init the default config for merging the raw config into it
        self.init_default_config()
        default_config_dict = self.dict()

        # default config that is updated by the raw config from the core
        raw_with_new_defaults = cast(
            Json,
            merge_json_elements(default_config_dict, raw_config_json),
        )

        # we also resolve the env var to construct a attrs class and not explode
        raw_with_resolved_env_vars = {
            k: replace_env_vars(v, os.environ, keep_unresolved=False) for k, v in raw_with_new_defaults.items()
        }

        # json with resolved env vars, defaults and dropped deleted attributes
        # this is the config that we will use to cleanup the raw_with_new_defaults
        reference_config_json = cast(
            Json,
            jsons.dump(
                Config.read_config(
                    raw_with_resolved_env_vars, reason="making reference config for cleanup"
                ),  # attrs class with dropped deleted attributes
                strip_attr="kind",
                strip_properties=True,
                strip_privates=True,
            ),
        )

        result = cast(Json, drop_deleted_attributes(raw_with_new_defaults, reference_config_json))
        return result

    @staticmethod
    def read_config(config: Json, read_as_json: bool = False, reason: Optional[str] = None) -> Dict[str, Any]:
        new_config = {}
        for config_id, config_data in config.items():
            if config_data is None:
                config_data = {}
            if config_id in Config.running_config.classes:
                message = f" reason: {reason}" if reason else ""
                log.debug(f"Loading config section {config_id}" + message)
                clazz: Type[Any] = Config.running_config.classes.get(config_id, Any) if not read_as_json else Any  # type: ignore # noqa: E501
                # use the from_json class from config, if available
                if loader := getattr(clazz, "from_json", None):
                    new_config[config_id] = loader(config_data)
                else:
                    new_config[config_id] = from_json(config_data, clazz)
            else:
                log.warning(f"Unknown config section {config_id}")
        return new_config

    @staticmethod
    def restart_required(new_config: Json) -> bool:
        for config_id, config_data in new_config.items():
            if config_id in Config.running_config.data:
                for field in fields(type(config_data)):
                    if field.metadata.get("restart_required", False):
                        old_value = getattr(Config.running_config.data[config_id], field.name, None)
                        new_value = getattr(config_data, field.name, None)
                        if new_value != old_value:
                            log.info(f"Changed config {config_id}.{field.name} requires restart")
                            return True
        return False

    @staticmethod
    def apply_path_overrides_resolve_env_vars(running_config: RunningConfig, config: Json) -> bool:
        # there was no config received from resotocore, keep defaults
        if not config:
            return False
        # resolve missing env_vars from the environment
        resolved_conf = {k: replace_env_vars(v, os.environ, keep_unresolved=False) for k, v in config.items()}

        running_config.data = Config.read_config(resolved_conf, reason="resolving env_vars on the final config")

        return True

    @staticmethod
    def override_config(running_config: RunningConfig) -> None:
        if getattr(ArgumentParser.args, "config_override", None) is None:
            return
        for override in getattr(ArgumentParser.args, "config_override", []):
            try:
                if "=" not in override:
                    log.error(f"Invalid config override {override}")
                    continue
                config_key, config_value = override.split("=", 1)
                if "." not in config_key:
                    log.error(f"Invalid config override {config_key}")
                    continue

                config_keys = config_key.split(".")
                num_keys = len(config_keys)
                config_part = running_config.data
                set_value = False

                # By default we cast the override value to the type of the current
                # value. This works for most cases including dictionary values.
                # Should the current value be None we see if there was a type specified
                # for the dataclass field and use it as a fallback.
                # This only works for dataclass fields.
                config_section = config_keys[0]
                top_config_key = config_keys[1]
                fallback_target_type = None
                if (
                    config_section in Config.running_config.types
                    and top_config_key in Config.running_config.types[config_section]
                ):
                    fallback_target_type = Config.running_config.types[config_section][top_config_key]

                for num_key, key in enumerate(config_keys):
                    if num_key == num_keys - 1:
                        set_value = True
                        log.debug(f"Overriding config key {config_key}")

                    if hasattr(config_part, key):
                        attr_value = getattr(config_part, key)
                        if set_value:
                            config_value = Config.cast_target_type(config_value, attr_value, fallback_target_type)
                            setattr(config_part, key, config_value)
                        else:
                            config_part = attr_value
                    elif isinstance(config_part, dict) and key in config_part:
                        attr_value = config_part[key]
                        if set_value:
                            config_value = Config.cast_target_type(config_value, attr_value, fallback_target_type)
                            config_part[key] = config_value
                        else:
                            config_part = attr_value
                    else:
                        log.error(f"Override key {config_key} is unknown - skipping")
                        break

                target_type = str
                if target_type in (list, tuple, set):
                    config_value = target_type(config_value.split(","))
                config_value = convert(config_value, target_type)

            except Exception:
                log.exception(f"Failed to override config {override}")

    @staticmethod
    def cast_target_type(config_value: Any, current_value: Any, fallback_target_type: Optional[type]) -> object:
        if current_value is None and fallback_target_type is not None:
            target_type = fallback_target_type
        else:
            target_type = type(current_value)
        if target_type in (list, tuple, set):
            config_value = target_type(config_value.split(","))
        else:
            config_value = convert(config_value, target_type)
        return config_value

    @staticmethod
    def dict() -> Json:
        return jsons.dump(Config.running_config.data, strip_attr="kind", strip_properties=True, strip_privates=True)  # type: ignore # noqa: E501

    def save_config(self) -> None:
        update_config_model(self.model, resotocore_uri=self.resotocore_uri, verify=self.verify)
        stored_config_revision = set_config(self.config_name, self.dict(), self.resotocore_uri, verify=self.verify)
        if stored_config_revision != Config.running_config.revision:
            Config.running_config.revision = stored_config_revision
            log.debug(f"Saved config {self.config_name} revision {Config.running_config.revision}")
        else:
            log.debug(f"Config {self.config_name} unchanged")

    def on_config_event(self, message: Dict[str, Any]) -> None:
        if (
            message.get("message_type") == "config-updated"
            and message.get("data", {}).get("id") == self.config_name
            and message.get("data", {}).get("revision") != Config.running_config.revision
        ):
            try:
                log.debug(f"Config {self.config_name} has changed - reloading")
                self.load_config(reload=True)
            except Exception:
                log.exception("Failed to reload config")

    # the __hash__ and the __eq__ below is a workaround to make sure the outdated config is not cached by
    # a lru_cache decoartor after the config performed a self-update. It serves no other purpose.
    def __hash__(self) -> int:
        return hash(self.running_config.revision)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Config):
            return self.running_config.__dict__ == other.running_config.__dict__
        return False

    @property
    def model(self) -> List[Json]:
        """Return the config dataclass model in resotocore format"""
        classes = set(_config.classes.values())
        return dataclasses_to_resotocore_model(classes)

    @staticmethod
    def add_args(arg_parser: ArgumentParser) -> None:
        arg_parser.add_argument(
            "--override",
            help="Override config attribute(s)",
            dest="config_override",
            type=str,
            default=[],
            nargs="+",
        )


# Note: the config is mutable.
def current_config() -> Config:
    # metaclass makes it possible to use the class as instance.
    # use this accessor here to get a typed instance of the config
    return Config  # type: ignore



from platform import python_implementation

import pytest
from resotolib.graph import Graph, GraphExportIterator, EdgeKey
from resotolib.baseresources import BaseResource, EdgeType, GraphRoot
import resotolib.logger as logger
from attrs import define
from typing import ClassVar

logger.getLogger("resoto").setLevel(logger.DEBUG)


@define(eq=False, slots=False)
class SomeTestResource(BaseResource):
    kind: ClassVar[str] = "some_test_resource"

    def delete(self, graph) -> bool:
        return False


def test_graph():
    g = Graph()
    n1 = SomeTestResource(id="foo", tags={})
    n2 = SomeTestResource(id="bar", tags={})
    g.add_node(n1)
    g.add_node(n2)
    g.add_edge(n1, n2)
    assert len(g.nodes) == 2
    assert len(g.edges) == 1


def test_graph_merge():
    rg1 = Graph()
    rg2 = Graph()
    a = SomeTestResource(id="a", tags={})
    b = SomeTestResource(id="b", tags={})
    c = SomeTestResource(id="c", tags={})
    d = SomeTestResource(id="d", tags={})
    rg1.add_node(a)
    rg1.add_node(b)
    rg2.add_node(c)
    rg2.add_node(d)
    rg1.add_edge(a, b, edge_type=EdgeType.delete)
    rg2.add_edge(c, d, edge_type=EdgeType.delete)
    rg1.merge(rg2)
    assert len(rg1.nodes) == 4
    assert len(rg1.edges) == 2
    for edge in rg1.edges:
        assert len(edge) == 3
        key = edge[2]
        assert len(key) == 3
        edge_type = key[2]
        assert edge_type == EdgeType.delete


def test_multidigraph():
    g = Graph()
    a = SomeTestResource(id="a", tags={})
    b = SomeTestResource(id="b", tags={})
    c = SomeTestResource(id="c", tags={})
    d = SomeTestResource(id="d", tags={})
    g.add_resource(a, b)
    g.add_resource(b, c)
    g.add_resource(c, d)
    g.add_edge(b, a, edge_type=EdgeType.delete)
    g.add_edge(b, d, edge_type=EdgeType.delete)
    assert len(g.nodes) == 4
    assert len(g.edges) == 5
    assert len(list(g.successors(a))) == 1
    g.add_edge(a, b)
    assert len(list(g.successors(a))) == 1
    assert len(list(g.predecessors(b))) == 1
    assert len(list(g.predecessors(a))) == 0
    assert len(list(g.successors(b))) == 1
    assert len(list(g.successors(c))) == 1
    assert len(list(g.successors(d))) == 0
    assert len(list(g.predecessors(a, edge_type=EdgeType.delete))) == 1
    assert len(list(g.successors(b, edge_type=EdgeType.delete))) == 2
    assert len(list(g.successors(a, edge_type=EdgeType.delete))) == 0
    assert len(list(g.predecessors(b, edge_type=EdgeType.delete))) == 0
    assert len(list(g.ancestors(a))) == 0
    assert len(list(g.descendants(a))) == 3
    assert len(list(g.descendants(a, edge_type=EdgeType.delete))) == 0
    assert len(list(g.descendants(b))) == 2
    assert len(list(g.descendants(b, edge_type=EdgeType.delete))) == 2
    assert g.is_acyclic_per_edge_type()
    g.add_edge(b, a)
    assert g.is_acyclic_per_edge_type() is False


# noinspection PyStatementEffect
def test_baseresource_chksum():
    g = Graph()
    a = SomeTestResource(id="a", tags={})
    with pytest.raises(RuntimeError):
        a.chksum
    g.add_node(a)
    assert isinstance(a.chksum, str)


@pytest.mark.skipif(condition=python_implementation() != "CPython", reason="not implemented")
def test_graph_export_iterator():
    from sys import getrefcount

    g = Graph(root=GraphRoot(id="root", tags={}))
    a = SomeTestResource(id="a", tags={})
    g.add_resource(g.root, a)
    assert getrefcount(g) == 2
    gei = GraphExportIterator(g)
    gei.export_graph()
    # one ref for g and for the variable passed to getrefcount ==> 2
    assert getrefcount(g) == 2
    assert len(list(gei)) == 3


def test_find_cycles():
    g = Graph()
    n1 = SomeTestResource(id="foo", tags={})
    n2 = SomeTestResource(id="bar", tags={})
    g.add_node(n1)
    g.add_node(n2)
    g.add_edge(n1, n2)
    assert g.is_acyclic_per_edge_type() is True
    g.add_edge(n2, n1)
    assert g.is_acyclic_per_edge_type() is False
    cycle = g.find_cycle()
    assert cycle == [
        EdgeKey(edge_type=EdgeType.default, src=n1, dst=n2),
        EdgeKey(edge_type=EdgeType.default, src=n2, dst=n1),
    ]



from __future__ import annotations

import json
import logging
from abc import ABC
from datetime import datetime
from enum import Enum
from typing import Dict, Any, Type, Union, Optional, Callable

from jsons import snakecase

from resotolib.types import Json, JsonElement
from resotolib.units import parse
from resotolib.utils import utc_str

log = logging.getLogger("resoto." + __name__)


# General idea and basic implementation is taken from: https://github.com/Onyo/jsonbender
class Bender(ABC):
    """
    Base bending class.
    """

    def __call__(self, source: Any) -> Any:
        return self.raw_execute(source).value

    def raw_execute(self, source: Any) -> Transport:
        transport = Transport.from_source(source)
        return Transport(self.execute(transport.value), transport.context)

    def execute(self, source: Any) -> Any:
        return source

    def or_else(self, other: Bender) -> Bender:
        return OrElse(self, other)

    def __eq__(self, other: Any) -> Bender:  # type: ignore
        return Eq(self, other)

    def __ne__(self, other: Any) -> Bender:  # type: ignore
        return Ne(self, other)

    def __and__(self, other: Any) -> Bender:
        return And(self, other)

    def __or__(self, other: Any) -> Bender:
        return Or(self, other)

    def __invert__(self: Any) -> Bender:
        return Invert(self)

    def __add__(self, other: Any) -> Bender:
        return Add(self, other)

    def __sub__(self, other: Any) -> Bender:
        return Sub(self, other)

    def __mul__(self, other: Any) -> Bender:
        return Mul(self, other)

    def __div__(self, other: Any) -> Bender:
        return Div(self, other)

    def __neg__(self) -> Bender:
        return Neg(self)

    def __truediv__(self, other: Any) -> Bender:
        return Div(self, other)

    def __floordiv__(self, other: Any) -> Bender:
        return Div(self, other)

    def __rshift__(self, other: Any) -> Bender:
        return Compose(self, other)

    def __lshift__(self, other: Any) -> Bender:
        return Compose(other, self)

    def __getitem__(self, index: Any) -> Bender:
        return self >> GetItem(index)


class BendingError(Exception):
    pass


Mapping = Union[Bender, Dict[str, Bender]]


class S(Bender):
    """
    Retrieve a value from a JSON object under given path.
    """

    def __init__(self, *path: str, default: Optional[Any] = None):
        if not path:
            raise ValueError("No path given")
        self._path = path
        self._default = default

    def execute(self, source: Any) -> Any:
        try:
            for key in self._path:
                source = source[key]
            return source
        except (KeyError, TypeError):
            return self._default


class K(Bender):
    """
    Selects a constant value.
    """

    def __init__(self, value: Any, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._val = value

    def execute(self, source: Any) -> Any:
        return self._val


class F(Bender):
    """
    Lifts a python callable into a Bender, so it can be composed.
    The extra positional and named parameters are passed to the function at
    bending time after the given value.

    Example:
    ```
    f = F(sorted, key=lambda d: d['id'])
    K([{'id': 3}, {'id': 1}]) >> f  #  -> [{'id': 1}, {'id': 3}]
    ```
    """

    def __init__(self, func: Callable[[Any], Any], *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._func = func
        self._args = args
        self._kwargs = kwargs

    def execute(self, value: Any) -> Any:
        # noinspection PyArgumentList
        return self._func(value, *self._args, **self._kwargs)


class OrElse(Bender):
    def __init__(self, source_bender: Bender, else_bender: Bender):
        super().__init__()
        self.source_bender = source_bender
        self.else_bender = else_bender

    def raw_execute(self, source: Any) -> Transport:
        first = self.source_bender.raw_execute(source)
        if first.value is not None:
            return first
        else:
            return self.else_bender.raw_execute(source)


class GetItem(Bender):
    """
    Can be applied to a list or dict bender via `[index]`.
    List: GetItem(0) -> first element
    Dictionary: GetItem('key') -> value of key
    """

    def __init__(self, index: Union[str, int]):
        self._index = index

    def execute(self, source: Any) -> Any:
        if isinstance(source, list) and isinstance(self._index, int):
            return source[self._index] if len(source) > abs(self._index) else None
        elif isinstance(source, dict) and isinstance(self._index, str):
            return source.get(self._index)
        else:
            return None


class Compose(Bender):
    """
    Compose two benders.
    Use `>>` instead of calling `Compose` directly.
    """

    def __init__(self, first: Bender, second: Bender):
        self._first = first
        self._second = second

    def raw_execute(self, source: Any) -> Transport:
        first = self._first.raw_execute(source)
        return self._second.raw_execute(first) if first.value is not None else first


class UnaryOperator(Bender):
    """
    Base class for unary bending operators.

    Whenever a unary op is activated, the op() method is called with the
    *value* (that is, the bender is implicitly activated).

    Subclasses must implement the op() method, which takes one value and
    should return the desired result.
    """

    def __init__(self, bender: Bender):
        self.bender = bender

    def op(self, v: Any) -> Any:
        raise NotImplementedError()

    def raw_execute(self, source: Any) -> Transport:
        source = Transport.from_source(source)
        val = self.op(self.bender(source))
        return Transport(val, source.context)


class Neg(UnaryOperator):
    def op(self, v: Any) -> Any:
        return -v


class Invert(UnaryOperator):
    def op(self, v: Any) -> Any:
        return not v


class BinaryOperator(Bender):
    """
    Base class for binary bending operators.

    Whenever a bin op is activated, the op() method is called with both
    *values* (that is, the benders are implicitly activated).

    Subclasses must implement the op() method, which takes two values and
    should return the desired result.
    """

    def __init__(self, bender1: Bender, bender2: Bender):
        self._bender1 = bender1
        self._bender2 = bender2

    def op(self, v1: Any, v2: Any) -> Any:
        raise NotImplementedError()

    def raw_execute(self, source: Any) -> Transport:
        source = Transport.from_source(source)
        val = self.op(self._bender1(source), self._bender2(source))
        return Transport(val, source.context)


class Add(BinaryOperator):
    def op(self, v1: Any, v2: Any) -> Any:
        return v1 + v2 if v1 is not None and v2 is not None else None


class Sub(BinaryOperator):
    def op(self, v1: Any, v2: Any) -> Any:
        return v1 - v2 if v1 is not None and v2 is not None else None


class Mul(BinaryOperator):
    def op(self, v1: Any, v2: Any) -> Any:
        return v1 * v2 if v1 is not None and v2 is not None else None


class Div(BinaryOperator):
    def op(self, v1: Any, v2: Any) -> Any:
        return float(v1) / float(v2) if v1 is not None and v2 is not None else None


class Eq(BinaryOperator):
    def op(self, v1: Any, v2: Any) -> Any:
        return v1 == v2


class Ne(BinaryOperator):
    def op(self, v1: Any, v2: Any) -> Any:
        return v1 != v2


class And(BinaryOperator):
    def op(self, v1: Any, v2: Any) -> Any:
        return v1 and v2


class Or(BinaryOperator):
    def op(self, v1: Any, v2: Any) -> Any:
        return v1 or v2


class Context(Bender):
    def raw_execute(self, source: Any) -> Transport:
        transport = Transport.from_source(source)
        return Transport(transport.context, transport.context)


class Transport:
    def __init__(self, value: Any, context: Dict[str, Any]):
        self.value = value
        self.context = context

    @classmethod
    def from_source(cls, source: Any) -> Transport:
        if isinstance(source, cls):
            return source
        else:
            return cls(source, {})


class Bend(Bender):
    def __init__(self, mappings: Mapping, **kwargs: Any):
        super().__init__(**kwargs)
        self._mappings = mappings

    def execute(self, value: Optional[Json]) -> Any:
        return bend(self._mappings, value) if value else None


class Sort(Bender):
    """
    Sort a list based on given extractor bender.
    """

    def __init__(self, extractor: Bender, **kwargs: Any):
        super().__init__(**kwargs)
        self._extractor = extractor

    def execute(self, source: Any) -> Any:
        if isinstance(source, list):
            return sorted(source, key=lambda x: bend(self._extractor, x))  # type: ignore
        else:
            return source


class AsInt(Bender):
    def execute(self, source: Any) -> Any:
        if isinstance(source, int):
            return source
        else:
            try:
                return int(source)
            except Exception:
                return None


class AsBool(Bender):
    def execute(self, source: Any) -> Any:
        if isinstance(source, bool):
            return source
        elif isinstance(source, str):
            return source.lower() in ("true", "yes", "1")
        else:
            return bool(source)


class ParseJson(Bender):
    def __init__(self, keys_to_snake: bool = False, **kwargs: Any):
        super().__init__(**kwargs)
        self._keys_to_snake = keys_to_snake

    def execute(self, source: Any) -> Any:
        def reformat_keys_to_snake(js: JsonElement) -> JsonElement:
            if isinstance(js, dict):
                return {snakecase(k): reformat_keys_to_snake(v) for k, v in js.items()}
            elif isinstance(js, list):
                return [reformat_keys_to_snake(v) for v in js]
            else:
                return js

        if isinstance(source, str):
            try:
                result = json.loads(source)
                return result if not self._keys_to_snake else reformat_keys_to_snake(result)
            except Exception:
                return None
        else:
            return None


class AsDate(Bender):
    """
    Parse a given input string as date.
    The format of the date needs to be defined.
    """

    def __init__(self, date_format: str = "%Y-%m-%dT%H:%M:%SZ", **kwargs: Any):
        super().__init__(**kwargs)
        self._format = date_format

    def execute(self, source: Any) -> Any:
        return datetime.strptime(source, self._format) if isinstance(source, str) else source


class ListOp(Bender, ABC):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._func: Callable[[Any], Any] = lambda x: x

    def op(self, func: Callable[[Any], Any], vals: Any) -> Any:
        raise NotImplementedError()

    def execute(self, source: Any) -> Any:
        return None if source is None else self.op(self._func, source)


class Forall(ListOp):
    """
    Similar to Python's map().
    Builds a new list by applying the given function to each element of the
    iterable.

    Example:
    ```
    Forall(lambda i: i * 2)(range(5))  # -> [0, 2, 4, 6, 8]
    ```
    """

    def op(self, func: Callable[[Any], Any], vals: Any) -> Any:
        return list(map(func, vals))

    @classmethod
    def bend(cls, mapping: Mapping, context: Optional[Dict[str, Any]] = None) -> Any:
        """
        Return a ForallBend instance that bends each element of the list with the
        given mapping.

        mapping: a JSONBender mapping as passed to the `bend()` function.
        context: optional. the context that will be passed to `bend()`.
                 Note that if context is not passed, it defaults at bend-time
                 to the one passed to the outer mapping.

        Example:
        ```
        source = [{'a': 23}, {'a': 27}]
        bender = Forall.bend({'b': S('a')})
        bender(source)  # -> [{'b': 23}, {'b': 27}]
        ```

        """
        return ForallBend(mapping, context)


class ForallBend(Forall):
    """
    Bends each element of the list with given mapping and context.

    mapping: a JSONBender mapping as passed to the `bend()` function.
    context: optional. the context that will be passed to `bend()`.
             Note that if context is not passed, it defaults at bend-time
             to the one passed to the outer mapping.
    """

    def __init__(self, mapping: Mapping, context: Optional[Dict[str, Any]] = None, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._mapping = mapping
        self._context = context

    def raw_execute(self, source: Any) -> Transport:
        transport = Transport.from_source(source)
        context = self._context or transport.context
        # ListOp.execute assumes the func is saved on self._func
        self._func = lambda v: bend(self._mapping, v, context)
        return Transport(self.execute(transport.value), transport.context)


class MapDict(Bender):
    """
    If you have a dict and want to map either key or value.
    """

    def __init__(self, key_bender: Optional[Bender] = None, value_bender: Optional[Bender] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self._key_bender = key_bender
        self._value_bender = value_bender

    def execute(self, value: Dict[Any, Any]) -> Dict[Any, Any]:
        def do_bend(v: Any, bender: Optional[Bender]) -> Any:
            return bender.raw_execute(v).value if bender else v

        return {do_bend(k, self._key_bender): do_bend(v, self._value_bender) for k, v in value.items()}


class StripNones(Bender):
    def execute(self, source: Any) -> Any:
        if isinstance(source, list):
            return [x for x in source if x is not None]
        else:
            return source


class MapValue(Bender):
    def __init__(self, lookup: Dict[str, Any], default: Any = None, **kwargs: Any):
        super().__init__(**kwargs)
        self._lookup = lookup
        self._default = default

    def execute(self, value: str) -> Any:
        return self._lookup.get(value, self._default)


class MapEnum(MapValue):
    def execute(self, value: str) -> Any:
        enum = super().execute(value)
        if isinstance(enum, Enum):
            return enum.value
        else:
            raise AttributeError(f"Mapping did not return an enumeration: in:{value} out:{enum}")


class StringToUnitNumber(Bender):
    def __init__(self, unit: str, expected: Type[Union[int, float]] = float, **kwargs: Any):
        super().__init__(**kwargs)
        self._unit = unit
        self._expected = expected

    def execute(self, value: str) -> Union[int, float]:
        return self._expected(parse(value).to(self._unit).magnitude)


class CPUCoresToNumber(Bender):
    def execute(self, source: str) -> float:
        return float(source[:-1]) / 1000 if isinstance(source, str) and source.endswith("m") else float(source)


class EmptyToNoneBender(Bender):
    def execute(self, source: Any) -> Any:
        return None if source in (None, "", [], {}) else source


class SecondsFromEpochToDatetime(Bender):
    def execute(self, source: int) -> str:
        return utc_str(datetime.utcfromtimestamp(source))


EmptyToNone = EmptyToNoneBender()


def bend(mapping: Mapping, source: Any, context: Optional[Dict[str, Any]] = None) -> Any:
    """
    The main bending function.

    mapping: the map of benders
    source: a dict to be bent

    returns a new dict according to the provided map.
    """

    def bend_with_context(inner: Mapping, transport: Transport) -> Any:
        if isinstance(inner, list):
            return [bend_with_context(v, transport) for v in inner]

        elif isinstance(inner, dict):
            res = {}
            for k, v in inner.items():
                try:
                    value = bend_with_context(v, transport)
                    res[k] = value
                except Exception as e:
                    log.error(e, exc_info=True)
                    m = "Error for key {}: {}".format(k, str(e))
                    raise BendingError(m)
            return res

        elif isinstance(inner, Bender):
            return inner(transport)

        else:
            return inner

    context = {} if context is None else context
    return bend_with_context(mapping, Transport(source, context))
